import os
import torch
from synthetic_dataset import generate_synthetic_data

from drn import *


def test_glm_mean():
    X_train, Y_train, _, _ = generate_synthetic_data()

    torch.manual_seed(1)
    glm = GLM(X_train.shape[1], distribution="gamma")

    # Either we call 'train' to move to same device as 'X_train', or do it manually
    glm = glm.to(X_train.device)

    glm.update_dispersion(X_train, Y_train)

    mean1 = glm.mean(X_train)
    mean2 = glm.distributions(X_train).mean
    assert torch.allclose(mean1, mean2)


def test_glm_save_load():
    X_train, Y_train, _, _ = generate_synthetic_data()

    torch.manual_seed(1)
    glm = GLM.from_statsmodels(X_train, Y_train, distribution="gamma")

    torch.save(glm.state_dict(), "glm.pt")

    glm_load = GLM(X_train.shape[1], distribution="gamma").to(X_train.device)
    glm_load.load_state_dict(torch.load("glm.pt"))

    os.remove("glm.pt")

    assert glm.state_dict().keys() == glm_load.state_dict().keys()
    for key in glm.state_dict().keys():
        assert torch.allclose(glm.state_dict()[key], glm_load.state_dict()[key])

    assert glm.dispersion == glm_load.dispersion



import copy
from typing import Callable, List, Union

import numpy as np
import pandas as pd

from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import StandardScaler
from sklearn.model_selection import train_test_split

import torch
import torch.nn as nn
from tqdm.auto import tqdm, trange

from .models import drn_cutpoints
from .models import GLM
from .models import DRN

# Define Criterion type hint
Criterion = Union[
    Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
    Callable[[List[torch.Tensor], torch.Tensor], torch.Tensor],
]


def train(
    model: nn.Module = None,
    criterion: Criterion = None,
    train_dataset: torch.utils.data.Dataset = None,
    val_dataset: torch.utils.data.Dataset = None,
    epochs=1000,
    patience=30,
    lr=0.001,
    device=None,
    log_interval=10,
    batch_size=None,
    optimizer=torch.optim.Adam,
    print_details=True,
    keep_best=True,
    gradient_clipping=False,
    num_hidden_layers=2,
    hidden_size=128,
    dropout_rate=0.2,
    baseline=None,
    cutpoints=None,
) -> None:
    """
    A generic training function given a model and a criterion & optimizer.
    """
    return_model = False
    if model is None:
        return_model = True
        # Extract input features (X_train) and target (y_train) from the training dataset
        X_train = train_dataset.tensors[0]

        # Initialize the DRN model
        torch.manual_seed(23)
        model = DRN(
            num_features=X_train.shape[1],
            cutpoints=cutpoints,
            glm=baseline,
            num_hidden_layers=num_hidden_layers,
            hidden_size=hidden_size,
            dropout_rate=dropout_rate,
        )

    if batch_size is None:
        batch_size = 128  # type: ignore
    train_loader = torch.utils.data.DataLoader(train_dataset, batch_size=batch_size)
    val_loader = torch.utils.data.DataLoader(val_dataset, batch_size=len(val_dataset))  # type: ignore

    if device is None:
        if torch.cuda.is_available():
            device = torch.device("cuda:0")
        # Temporarily disable MPS which doesn't support some operations we need.
        # elif torch.backends.mps.is_available():
        #     device = torch.device("mps")
        else:
            device = torch.device("cpu")

    try:
        if print_details:
            print(f"Using device: {device}")
        model.to(device)

        optimizer = optimizer(model.parameters(), lr=lr)

        no_improvement = 0
        best_loss = torch.Tensor([float("inf")]).to(device)
        best_model = model.state_dict()

        range_selection = trange if print_details else range
        for epoch in range_selection(1, epochs + 1):
            model.train()

            for data, target in train_loader:
                x = data.to(device)
                y = target.to(device)
                optimizer.zero_grad()
                loss = criterion(model(x), y)
                loss.backward()

                # Check for NaN gradients
                for name, param in model.named_parameters():
                    if (
                        param.grad is not None
                    ):  # Parameters might not have gradients if they are not trainable
                        if torch.isnan(param.grad).any():
                            if print_details:
                                tqdm.write("Stopping training as NaN gradient detected")
                            raise ValueError(
                                f"Gradient NaN detected in {name}. Try smaller learning rates."
                            )

                if gradient_clipping:
                    # If no error is raised, proceed with gradient clipping and optimizer step
                    torch.nn.utils.clip_grad_norm_(model.parameters(), max_norm=1.0)

                optimizer.step()

            model.eval()

            loss_val = torch.zeros(1, device=device)

            for data, target in val_loader:
                x = data.to(device)
                y = target.to(device)

                loss_val += criterion(model(x), y)

            if loss_val < best_loss:
                best_loss = loss_val
                best_model = copy.deepcopy(model.state_dict())
                no_improvement = 0

            else:
                no_improvement += 1

            if print_details and epoch % log_interval == 0:
                tqdm.write(
                    f"Epoch {epoch} \t| batch train loss: {loss.item():.4f}"
                    + f"\t| validation loss:  {loss_val.item():.4f}"
                    + f"\t| no improvement: {no_improvement}"
                )

            if no_improvement > patience:
                if print_details:
                    tqdm.write("Stopping early!")
                break

    finally:
        if keep_best:
            # Load the best model found during training
            model.load_state_dict(best_model)

        # Make sure dropout is always disabled after training
        model.eval()

        return model


def split_and_preprocess(
    features, target, num_features, cat_features, seed=42, num_standard=True
):
    # Before preprocessing split
    x_train_raw, x_test_raw, y_train, y_test = train_test_split(
        features, target, random_state=seed, train_size=0.8, shuffle=True
    )

    x_train_raw, x_val_raw, y_train, y_val = train_test_split(
        x_train_raw, y_train, random_state=seed, train_size=0.75, shuffle=True
    )

    # Determine the full set of categories for each feature
    all_categories = {feature: set() for feature in cat_features}
    for feature in cat_features:
        all_categories[feature].update(features[feature].unique())

    # Convert each categorical feature to a categorical type with all possible categories
    for feature in cat_features:
        # Sort the categories to ensure consistent order
        sorted_categories = sorted(all_categories[feature])
        x_train_raw[feature] = pd.Categorical(
            x_train_raw[feature], categories=sorted_categories
        )
        x_val_raw[feature] = pd.Categorical(
            x_val_raw[feature], categories=sorted_categories
        )
        x_test_raw[feature] = pd.Categorical(
            x_test_raw[feature], categories=sorted_categories
        )
        features[feature] = pd.Categorical(
            features[feature], categories=sorted_categories
        )

    # One-hot Encoding
    features_one_hot = pd.get_dummies(features, columns=cat_features)
    features_one_hot = features_one_hot.astype(float)

    x_train, x_test, y_train, y_test = train_test_split(
        features_one_hot, target, random_state=seed, train_size=0.8, shuffle=True
    )

    x_train, x_val, y_train, y_val = train_test_split(
        x_train, y_train, random_state=seed, train_size=0.75, shuffle=True
    )

    if num_standard:
        # Standarise the numeric features.
        ct = ColumnTransformer(
            [("standardize", StandardScaler(), num_features)],
            remainder="passthrough",
            verbose_feature_names_out=False,
        )

        x_train = ct.fit_transform(x_train)
        x_val = ct.transform(x_val)
        x_test = ct.transform(x_test)
    else:
        ct = None

    x_train = pd.DataFrame(
        x_train, index=x_train_raw.index, columns=features_one_hot.columns
    )
    x_val = pd.DataFrame(x_val, index=x_val_raw.index, columns=features_one_hot.columns)
    x_test = pd.DataFrame(
        x_test, index=x_test_raw.index, columns=features_one_hot.columns
    )

    return (
        x_train,
        x_val,
        x_test,
        y_train,
        y_val,
        y_test,
        x_train_raw,
        x_val_raw,
        x_test_raw,
        num_features,
        cat_features,
        all_categories,
        ct,
    )



import re

import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
from matplotlib import ticker

import numpy as np
import pandas as pd
import seaborn as sns
import shap
import torch
import sklearn
from scipy.stats import gaussian_kde
from sklearn.compose import ColumnTransformer


class KernelSHAP_DRN:
    """
    This class produces the Kernel SHAP values regarding the distributional property of interest.
    It produces the raw Kernel SHAP values.
    It also generates SHAP dependence plot for any pair of features, considering categorical features.
    Beeswarm plot can be generated for any features.
    """

    def __init__(
        self,
        explaining_data,
        nsamples_background_fraction,
        background_data_df_before_one_hot: pd.DataFrame,
        one_hot_encoder,
        value_function,
        glm_value_function,
        other_shap_values=None,
        random_state=42,
    ):
        """
        Args:
        See the DRNExplainer class for explanations regarding
        {explaining_data, nsamples_background_fraction, background_data_df_before_one_hot, one_hot_encoder}
        value_function: v_{M}(S, x), given any instance x and indices S \\subseteq \\{1, ..., p\\}
        """
        super(KernelSHAP_DRN, self).__init__()
        self.background_data_df_before_one_hot = background_data_df_before_one_hot
        self.value_function = value_function
        self.explaining_data = explaining_data
        self.other_shap_values = other_shap_values
        self.feature_names = self.background_data_df_before_one_hot.columns
        self.one_hot_encoder = one_hot_encoder
        sample_size = int(
            np.round(
                self.background_data_df_before_one_hot.shape[0]
                * nsamples_background_fraction
            )
        )

        if self.other_shap_values is None:
            # Compute SHAP values for the DRN network
            np.random.seed(random_state)
            kernel_shap_explainer = shap.KernelExplainer(
                self.value_function,
                shap.sample(
                    self.background_data_df_before_one_hot,
                    nsamples=sample_size,
                    random_state=random_state,
                ),
            )
            self.shap_values_kernel = kernel_shap_explainer(
                self.explaining_data,
            )
            self.shap_base_values = self.shap_values_kernel.base_values
            self.shap_values = self.shap_values_kernel.values

            # Compute SHAP values for the GLM if required
            np.random.seed(random_state)
            self.glm_value_function = glm_value_function
            if self.glm_value_function is not None:
                kernel_shap_explainer_glm = shap.KernelExplainer(
                    self.glm_value_function,
                    shap.sample(
                        self.background_data_df_before_one_hot,
                        nsamples=sample_size,
                        random_state=random_state,
                    ),
                )
                self.shap_values_kernel_glm = kernel_shap_explainer_glm(
                    self.explaining_data,
                )

    def forward(self):
        """
        The raw Kernel SHAP (either adjusted or DRN) output.
        """
        return self.shap_values_kernel

    def shap_glm_values(self):
        """
        The raw Kernel SHAP (GLM) output.
        """
        return self.shap_values_kernel_glm

    def shap_values_mean_adjustments(self):
        """
        The SHAP values and feature names
        """
        return (self.shap_base_values, self.shap_values, self.feature_names)

    def shap_dependence_plot(self, features_tuple, output="value"):
        """
        Create the SHAP dependence plots
        features_tuple: the pair of features required for plotting
        other_shap_values: allows for externally calculated SHAP values, i.e., FastSHAP...
        """
        tuple_indexes = [
            np.where(self.feature_names == feature)[0][0]
            for feature in features_tuple
            if feature in self.feature_names
        ]
        instances = self.explaining_data.copy()

        # Convert categorical features to numeric
        #  Encoding details
        encoders = {}
        for feature in features_tuple:
            if (
                instances[feature].dtype == object
                or instances[feature].dtype == "category"
            ):
                # not isinstance(instances[feature].values[0], float):
                encoder = sklearn.preprocessing.LabelEncoder()
                instances[feature] = encoder.fit_transform(instances[feature])
                encoders[feature] = encoder
                for class_label, encoding in zip(
                    encoder.classes_, encoder.transform(encoder.classes_)
                ):
                    print(f"{feature}: {class_label} -> {encoding}")

        if self.other_shap_values is not None:
            shap_for_tuple = self.other_shap_values
        else:
            shap_for_tuple = self.shap_value_selection(tuple_indexes, output)

        instances_for_tuple = instances.values[:, tuple_indexes]
        feature_name_tuple = self.feature_names[tuple_indexes]

        batch_size = shap_for_tuple.shape[0]

        fig, axes = plt.subplots(1, 2, figsize=(25, 10))  # 1 row, 2 columns

        for feature_idx in range(2):
            shap_values = shap_for_tuple[:, feature_idx : (feature_idx + 1)].reshape(
                batch_size, 1
            )
            feature_values = instances_for_tuple[:, feature_idx : (feature_idx + 1)]

            # Determine the color based on the other feature's value
            color_feature_index = 1 - feature_idx
            colors = instances_for_tuple[
                :, color_feature_index : (color_feature_index + 1)
            ]

            ax = axes[feature_idx]
            scatter = ax.scatter(
                feature_values, shap_values, alpha=0.85, c=colors, cmap="viridis", s=22
            )

            # Handle categorical feature for x-axis
            if feature_name_tuple[feature_idx] in encoders:
                encoder = encoders[feature_name_tuple[feature_idx]]
                ticks_and_labels = list(enumerate(encoder.classes_))
                ticks, labels = zip(*ticks_and_labels)
                ax.set_xticks(ticks)
                ax.set_xticklabels(labels)

            # Improved Handling of Categorical Feature for Color Bar
            if feature_name_tuple[color_feature_index] in encoders:
                encoder = encoders[feature_name_tuple[color_feature_index]]
                # Create a color map with a color for each unique value in the feature
                unique_vals = np.unique(instances_for_tuple[:, color_feature_index])
                cmap = mcolors.ListedColormap(
                    plt.cm.viridis(np.linspace(0, 1, len(unique_vals)))
                )
                norm = mcolors.BoundaryNorm(
                    np.arange(-0.5, len(unique_vals) + 0.5, 1), cmap.N
                )

                scatter.set_cmap(cmap)
                scatter.set_norm(norm)

                # Create the colorbar using the specified cmap and norm
                cbar = plt.colorbar(scatter, ax=ax, ticks=np.arange(len(unique_vals)))
                cbar.set_ticklabels(encoder.classes_)
                cbar.set_label(
                    f"{feature_name_tuple[color_feature_index]}", fontsize=30
                )

                # Ensuring dot-like representation for categorical color bar
                cbar.ax.minorticks_off()  # This removes any minor ticks
                cbar.ax.get_yaxis().set_major_locator(ticker.MaxNLocator(integer=True))
            else:
                cbar = plt.colorbar(scatter, ax=ax)
                cbar.set_label(
                    f"{feature_name_tuple[color_feature_index]}", fontsize=30
                )  # Set the font size as needed

            ax.set_xlabel(f"{feature_name_tuple[feature_idx]}", fontsize=30)
            ax.set_ylabel(f"$\\phi_{{{tuple_indexes[feature_idx]+1}}}$", fontsize=30)
            ax.set_title(
                f"SHAP Values for {feature_name_tuple[feature_idx]}", fontsize=30
            )

        # plt.tick_params(axis='x', labelsize=14)
        # plt.tick_params(axis='y', labelsize=14)
        plt.tight_layout()

    def beeswarm_plot(self, features=None, output="value"):
        """
        Create the beeswarm summary plots
        features: a list of feature names required for plotting
        adjusting: False --> explaining the drn model; True --> explaining how the drn adjusts the glm
        """
        features = (
            self.background_data_df_before_one_hot.columns
            if features is None
            else features
        )
        features_idx = [
            np.where(self.feature_names == feature)[0][0]
            for feature in features
            if feature in self.feature_names
        ]

        if self.other_shap_values is not None:
            shap_values = self.other_shap_values
        else:
            shap_values = self.shap_value_selection(features_idx, output)

        # Beeswarm summary plot
        shap.summary_plot(
            shap_values, self.explaining_data.iloc[:, features_idx], plot_size=(8, 6)
        )

    def global_importance_plot(self, features=None, output="value"):
        """
        Creates a global importance plot based on the absolute SHAP values.
        """
        features = (
            self.background_data_df_before_one_hot.columns
            if features is None
            else features
        )
        features_idx = [
            np.where(self.feature_names == feature)[0][0]
            for feature in features
            if feature in self.feature_names
        ]

        if self.other_shap_values is not None:
            shap_values = self.other_shap_values
        else:
            shap_values = self.shap_value_selection(features_idx, output)

        feature_names = self.feature_names

        # Sum the absolute SHAP values for each feature across all samples
        shap_sum = np.abs(shap_values).mean(axis=0)

        # Sort the features by their importance
        feature_importance = sorted(
            zip(feature_names, shap_sum), key=lambda x: x[1], reverse=True
        )
        sorted_features, sorted_importances = zip(*feature_importance)

        # Create the plot
        plt.figure(figsize=(15, 15))
        plt.barh(
            range(len(sorted_importances)),
            sorted_importances,
            tick_label=sorted_features,
        )
        plt.xlabel("Mean Absolute SHAP Value (Global Importance)", fontsize=35)
        plt.ylabel("Features", fontsize=35)
        plt.title("Global Feature Importance Based on SHAP Values", fontsize=35)
        plt.gca().invert_yaxis()  # Invert y-axis to have the most important feature on top

    def shap_value_selection(self, indexes, output="value"):
        if output == "value":
            shap_values = self.shap_values[:, indexes]
        elif output == "glm":
            if self.glm_value_function is None:
                raise ValueError(f"Set glm_output = True while initilising the class!")
            else:
                shap_values = self.shap_values_kernel_glm.values[:, indexes]
        elif output == "drn":
            if self.glm_value_function is None:
                raise ValueError(f"Set glm_output = True while initilising the class!")
            else:
                shap_values = (
                    self.shap_values[:, indexes]
                    + self.shap_values_kernel_glm.values[:, indexes]
                )
        return shap_values


class DRNExplainer:
    def __init__(
        self,
        drn: torch.nn.Module,
        glm: torch.nn.Module,
        default_cutpoints: list,
        background_data_df_before_one_hot: pd.DataFrame,
        cat_features: list,
        all_categories: list = None,
        column_transformer: ColumnTransformer = None,
    ):
        """
        Initialise the DRNExplainer with the given parameters.

        Args:
            drn (torch.nn.Module): The DRN neural network model.
            glm (torch.nn.Module): The baseline Generalised Linear Model (GLM).
            default_cutpoints (list): Cutpoints used for training the DRN.
            background_data_df_before_one_hot (pd.DataFrame): Background data prior to one-hot encoding.
            cat_features (list): List of categorical features.
            all_categories (list, optional): List of all categories of the categorical features, if available.
            column_transformer (ColumnTransformer, optional): Preprocessor for numerical features.
        """

        super().__init__()  # Initialize the superclass. Needed if the superclass has an __init__ to run.

        # Assign the provided arguments to instance variables.
        self.drn = drn  # The DRN model.
        self.glm = glm  # The baseline GLM model.
        self.default_cutpoints = (
            default_cutpoints  # Cutpoints for feature partitioning.
        )
        self.column_transformer = (
            column_transformer  # Transformer for scaling numerical features.
        )

        # Store the raw data and features information.
        self.background_data_df_before_one_hot = (
            background_data_df_before_one_hot  # Data before one-hot encoding.
        )
        self.cat_features = cat_features  # Categorical feature names.
        self.all_categories = (
            all_categories  # Categories for all features if available.
        )

        # One-hot encode background data if categories are provided, else store it as is.
        self.background_data_df = (
            self.one_hot_encoder(self.background_data_df_before_one_hot)
            if self.all_categories is not None
            else self.background_data_df_before_one_hot
        )

        self.background_data_tensor = self._to_tensor(
            self.background_data_df
        )  # Convert the data to a PyTorch tensor.

    def plot_dp_adjustment_shap(
        self,
        instance_raw: pd.DataFrame,
        dist_property: str = "Mean",
        quantile_bounds: tuple = None,
        method: str = "Kernel",
        nsamples_background_fraction: float = 1.0,
        top_K_features: int = 3,
        adjustment: bool = True,
        other_df_models: list = None,
        model_names: list = None,
        cutpoints: list = None,
        num_interpolations: int = None,
        labelling_gap: float = 0.05,
        synthetic_data=None,
        synthetic_data_samples: int = int(1e6),
        observation=None,
        plot_baseline: bool = True,
        # Plotting parameters below:
        x_range: tuple = None,
        y_range: tuple = None,
        plot_y_label: str = None,
        plot_title: str = None,
        figsize=None,
        density_transparency: float = 1.0,
        shap_fontsize: int = 25,
        legend_loc: str = "upper left",
    ):
        """
        Plot SHAP value-based adjustments with an option to include density functions.

        Args:
            instance_raw: Raw data before one-hot encoding, used for instance-specific analysis.
            dist_property: Distributional property to adjust ('Mean', 'Variance', 'Quantile').
            method: Technique for SHAP value computation ('Kernel', 'Tree', 'FastSHAP', etc.).
            nsamples_background_fraction: Fraction of background data for SHAP calculation.
            top_K_features: Number of top features based on SHAP values.
            adjustment: Whether to plot the SHAP values of the adjusted or unadjusted distributional property
            other_df_models: Other distributional forecasting models for comparison.
            model_names: Names of the other distributional forecasting models.
            cutpoints: Cutpoints for partitioning feature space, defaults to `self.default_cutpoints`.
            num_interpolations: Number of points for density interpolation, defaults to 2000.
            labelling_gap: Gap between labels in the plot for readability.
            synthetic_data, synthetic_data_samples: Synthetic data function for true density comparison and number of samples generated.
            observation: Specific observation value for vertical line plotting.
            plot_baseline: Flag to include baseline model's density plot.
            x_range, y_range: Axis ranges for the plot.
            plot_y_label, plot_title: Custom labels for the plot's axes and title.
            density_transparency: Alpha value for density plot transparency.
            shap_fontsize, figsize, label_adjustment_factor: Plot styling parameters.
            legend_loc: Location of the legend in the plot.
        """
        alpha = density_transparency

        # Prepare data for plotting: One-hot encode the input and convert to a tensor.
        print("f", instance_raw)
        instance = self._to_tensor(self.one_hot_encoder(instance_raw))

        # Use default cutpoints unless explicitly provided.
        cutpoints = self.default_cutpoints if cutpoints is None else cutpoints

        # Set number of interpolation points for density estimation; default is 2000.
        num_interpolations = 2000 if num_interpolations is None else num_interpolations

        # Determine the range for density interpolation based on either specified x_range or the given cutpoints.
        lower_cutpoint = x_range[0] if x_range is not None else cutpoints[0]
        upper_cutpoint = x_range[1] if x_range is not None else cutpoints[-1]
        # Create a tensor of linearly spaced values between the lower and upper cutpoints.
        x_grid = torch.linspace(
            lower_cutpoint, upper_cutpoint, num_interpolations
        ).unsqueeze(-1)

        # Compute probability density functions (PDF) for GLM and DRN models.
        # GLM PDF is calculated from the log probabilities of the distribution.
        glm_pdf = np.exp(
            (self.glm.distributions(instance).log_prob(x_grid)).detach().numpy()
        )

        # DRN PDF is calculated from the probabilities of the distribution.
        drn_pdf = self.drn.distributions(instance).prob(x_grid).detach().numpy()

        # Setup the plotting environment with specified figure size.
        figure, axes = plt.subplots(1, 1, figsize=figsize)
        if axes is not None and figsize is not None:
            axes.figure.set_size_inches(figsize)
        if axes is None:
            _, axes = plt.subplots(1, 1, figsize=figsize)

        plt.sca(axes)

        # Calculate the maximum Y-axis value for the plot
        y_max = 1.25 * max(drn_pdf.max(), glm_pdf[1:].max())
        if y_range is not None:
            y_min, y_max = y_range

        # Define the X-axis range
        x_min, x_max = 0, cutpoints[-1]

        # Find indices of cutpoints closest to the default cutpoints
        index_right = torch.argmin(torch.abs(x_grid - upper_cutpoint)).item()
        index_left = torch.argmin(torch.abs(x_grid - lower_cutpoint)).item()

        # Baseline Density Plotting
        if plot_baseline:
            plt.plot(
                x_grid[index_left:index_right],
                glm_pdf[index_left:index_right],
                color="black",
                linewidth=3,
                label="GLM",
                alpha=alpha,
            )

        plot_drn_density(
            x_grid,
            drn_pdf,
            index_left,
            index_right,
            axes,
            alpha,
        )

        # Integrate SHAP explanations into the plot
        self.kernel_shap_plot(
            instance_raw=instance_raw,
            instance=instance,
            dist_property=dist_property,
            quantile_bounds=quantile_bounds,
            method=method,
            nsamples_background_fraction=nsamples_background_fraction,
            adjustment=adjustment,
            axes=axes,
            top_K_features=top_K_features,
            y_max=y_max,
            y_min=y_min,
            labelling_gap=labelling_gap,
            fontsize=shap_fontsize,
        )

        # Plot synthetic data density or vertical lines to denote true values if synthetic data is provided
        if synthetic_data is not None:
            instance_np = instance.detach().numpy()[0]
            y = np.array(
                synthetic_data(
                    synthetic_data_samples, 0, specific_instance=instance_np
                )[1].values
            )
            print(instance, instance_np)
            if observation:
                # Plot vertical line for the observed property mean or quantile
                property_value = (
                    np.mean(y) if dist_property == "Mean" else np.percentile(y, 90)
                )
                axes.axvline(
                    property_value,
                    ls="dashed",
                    alpha=0.75,
                    linewidth=4,
                    label=f"True {dist_property}: {round(float(property_value), 3)}",
                    color="orange",
                )
            else:
                # Plot KDE for the synthetic data
                sns.kdeplot(
                    y,
                    color="orange",
                    label="True Density",
                    gridsize=3000,
                    linewidth=3,
                    alpha=alpha,
                )

        # Plot observed value if synthetic data is not available
        if observation is not None and synthetic_data is None:
            axes.axvline(
                observation,
                ls="dashed",
                alpha=0.75,
                linewidth=4,
                label=f"Observation:\n{round(float(observation), 3)}",
                color="red",
            )

        # Plot density functions from other distribution forecasting models
        if other_df_models is not None:
            for idx, DF_model in enumerate(other_df_models):
                current_pdf = np.exp(
                    DF_model.distributions(instance).log_prob(x_grid).detach().numpy()
                )
                plt.plot(
                    x_grid,
                    current_pdf,
                    label=f"{model_names[idx]} Density",
                    color=str(0.5 + 0.5 * idx / len(other_df_models)),
                    alpha=alpha,
                )

        # Set axis labels, title, and legend
        plt.xlabel("Y")
        plt.gca().set_ylabel("")
        if plot_y_label is not None:
            plt.ylabel(plot_y_label)
        if plot_title is not None:
            plt.title(plot_title)
        else:
            plt.title(
                f'{dist_property} {"Adjustment" if adjustment else "Explanation"}',
            )

        plt.legend(loc=legend_loc)

        # Set plot ranges
        if x_range is not None:
            x_min, x_max = x_range
        axes.set_xlim(x_min, x_max)
        axes.set_ylim(0, y_max)

    def plot_adjustment_factors(
        self,
        instance,
        observation=None,
        cutpoints=None,
        num_interpolations=None,
        other_df_models=None,
        model_names=None,
        percentiles=None,
        cutpoints_label_bool=False,
        synthetic_data=None,
        plot_adjustments_labels=True,
        axes: plt.Axes = None,
        x_range=None,
        y_range=None,
        plot_title=None,
        plot_mean_adjustment=False,
        plot_y_label=None,
        density_transparency=1.0,
        figsize=None,
    ):
        """
        Plot the adjustment factors for each of the partitioned interval.
        expand: interpolation of cutpoints for density evaluations.
        """
        alpha = density_transparency

        # Default cutpoints and num_interpolations setup
        cutpoints = self.default_cutpoints if cutpoints is None else cutpoints
        c_0 = self.default_cutpoints[0]
        c_K = self.default_cutpoints[-1]
        instance = (
            self._to_tensor(self.one_hot_encoder(instance))
            if self.all_categories is not None
            else self._to_tensor(instance)
        )

        # If we are plotting concerning quantiles
        if percentiles is not None:
            cutpoints_label_bool = True
            cutpoints = (
                self.glm.quantiles(instance, percentiles).detach().numpy().squeeze(1)
            )

            cutpoints[0] = cutpoints[0] + 1e-3

        # Interpolation
        lower_cutpoint = x_range[0] if x_range is not None else cutpoints[0]
        upper_cutpoint = x_range[1] if x_range is not None else cutpoints[-1]
        x_grid = torch.linspace(
            lower_cutpoint, upper_cutpoint, num_interpolations
        ).unsqueeze(-1)

        # Calculate GLM and DRN PDFs
        glm_pdf = np.exp(
            self.glm.distributions(instance).log_prob(x_grid).detach().numpy()
        )
        drn_pdf = self.drn.distributions(instance).prob(x_grid).detach().numpy()

        # Set up the plot
        if axes is not None and figsize is not None:
            axes.figure.set_size_inches(figsize)
        if axes is None:
            _, axes = plt.subplots(1, 1, figsize=figsize)

        plt.sca(axes)

        y_max = 1.5 * max(drn_pdf.max(), glm_pdf[1:].max())
        x_min, x_max = cutpoints[0], cutpoints[-1]

        # Find indices of cutpoints closest to the default cutpoints
        index_right = torch.argmin(torch.abs(x_grid - upper_cutpoint)).item()
        index_left = torch.argmin(torch.abs(x_grid - lower_cutpoint)).item()

        # Baseline Density Plotting
        plt.plot(x_grid, glm_pdf, color="blue", alpha=alpha)
        plt.plot(
            x_grid[index_left:index_right],
            glm_pdf[index_left:index_right],
            color="black",
            linewidth=2,
            label="GLM",
            alpha=alpha,
        )

        plot_drn_density(
            x_grid,
            drn_pdf,
            index_left,
            index_right,
            axes,
            alpha,
        )

        # Other DF Models Plotting
        if other_df_models is not None:
            for idx, DF_model in enumerate(other_df_models):
                current_pdf = np.exp(
                    DF_model.distributions(instance).log_prob(x_grid).detach().numpy()
                )
                plt.plot(
                    x_grid,
                    current_pdf,
                    label=f"{model_names[idx]} Density",
                    color=str(0.5 + 0.5 * idx / len(other_df_models)),
                    alpha=alpha,
                )

        if plot_adjustments_labels:
            if cutpoints_label_bool:
                for i in range(len(cutpoints)):
                    if cutpoints[i] <= c_0 and cutpoints[i + 1] > c_0:
                        plt.text(
                            c_0,
                            y_max / 35,
                            "Adjusted $c_{0}$",
                            ha="center",
                            fontsize=14,
                            color="black",
                        )
                        plt.axvline(c_0, ls="--", color="black", alpha=0.5)
                    elif (
                        cutpoints[i] >= x_grid[index_left]
                        and cutpoints[i] <= x_grid[index_right]
                    ):
                        plt.text(
                            cutpoints[i],
                            y_max * 0.8,
                            f"Baseline\n {percentiles[i]}% Quantile",
                            ha="center",
                            fontsize=20,
                            color="black",
                        )
                        plt.axvline(cutpoints[i], ls="--", color="black", alpha=0.5)
                    elif cutpoints[i] >= c_K:
                        plt.text(
                            c_K,
                            y_max / 35,
                            "Adjusted $c_{K}$",
                            ha="center",
                            fontsize=14,
                            color="black",
                        )
                        plt.axvline(c_K, ls="--", color="black", alpha=0.5)
                        break

            interval_width = int(num_interpolations / len(cutpoints))
            adjustment_idx = 1
            for i in range(len(cutpoints) - 1):
                if (
                    cutpoints[-1] <= self.default_cutpoints[0]
                    or cutpoints[0] >= self.default_cutpoints[-1]
                ):
                    region_start = cutpoints[0]
                    region_end = cutpoints[-1]
                    if not cutpoints_label_bool:
                        plt.axvline(region_end, ls="--", color="black", alpha=0.5)

                    plt.text(
                        (region_start + region_end) / 2,
                        y_max / 2,
                        f"No Adjustment",
                        ha="center",
                        fontsize=16,
                        color="black",
                    )
                    break

                # cutpoint_{i} < c_0 < cutpoint_{i +1}
                elif (
                    cutpoints[i] < self.default_cutpoints[0]
                    and cutpoints[i + 1] >= self.default_cutpoints[0]
                ):
                    # we need to break it down into [cutpoint_0, c_0) and [c_0, cutpoint_{i+1})
                    # [cutpoint_0, c_0) should be labelled as "no adjustment"
                    region_start = cutpoints[0]
                    region_end = self.default_cutpoints[0]
                    if not cutpoints_label_bool:
                        plt.axvline(region_end, ls="--", color="black", alpha=0.5)
                    plt.text(
                        (region_start + region_end) / 2,
                        y_max / 2,
                        f"No Adjustment",
                        ha="center",
                        fontsize=16,
                        color="black",
                    )

                    if cutpoints[i + 1] <= self.default_cutpoints[-1]:
                        # [c_0, cutpoint_{i+1})
                        adjustment_idx = self.region_text(
                            instance,
                            interval_width,
                            drn_pdf,
                            glm_pdf,
                            y_max,
                            region_start=self.default_cutpoints[0],
                            region_end=cutpoints[i + 1],
                            cutpoint_idx=i,
                            adjustment_idx=1,
                            cutpoints_label_bool=cutpoints_label_bool,
                            percentiles=percentiles,
                        )
                    else:
                        region_start = self.default_cutpoints[0]
                        region_end = self.default_cutpoints[-1]
                        if not cutpoints_label_bool:
                            plt.axvline(region_end, ls="--", color="black", alpha=0.5)
                        plt.text(
                            (region_start + region_end) / 2,
                            y_max / 2,
                            f"No Adjustment",
                            ha="center",
                            fontsize=16,
                            color="black",
                        )

                        region_start = self.default_cutpoints[-1]
                        region_end = cutpoints[i + 1]
                        plt.text(
                            (region_start + region_end) / 2,
                            y_max / 2,
                            f"No Adjustment",
                            ha="center",
                            fontsize=16,
                            color="black",
                        )
                        break

                # cutpoint_{i} < c_K <= cutpoints_{i+1}
                elif (
                    cutpoints[i + 1] >= self.default_cutpoints[-1]
                    and cutpoints[i] >= self.default_cutpoints[0]
                ):
                    # we need to break it down into [cutpoint_{i}, c_K) and [c_K, cutpoint_{i+1})
                    adjustment_idx = self.region_text(
                        instance,
                        interval_width,
                        drn_pdf,
                        glm_pdf,
                        y_max,
                        cutpoints[i],
                        self.default_cutpoints[-1],
                        i,
                        adjustment_idx,
                        cutpoints_label_bool=cutpoints_label_bool,
                        percentiles=percentiles,
                    )

                    region_start = self.default_cutpoints[-1]
                    region_end = cutpoints[-1]
                    if region_end != region_start:
                        plt.text(
                            (region_start + region_end) / 2,
                            y_max / 2,
                            f"No Adjustment",
                            ha="center",
                            fontsize=16,
                            color="black",
                        )
                    break

                # c_0 < cutpoint_{i} < cutpoints_{i+1} < c_K
                elif (
                    cutpoints[i + 1] < self.default_cutpoints[-1]
                    and cutpoints[i] >= self.default_cutpoints[0]
                    and cutpoints[i] < self.default_cutpoints[-1]
                ):

                    adjustment_idx = self.region_text(
                        instance,
                        interval_width,
                        drn_pdf,
                        glm_pdf,
                        y_max,
                        cutpoints[i],
                        cutpoints[i + 1],
                        i,
                        adjustment_idx,
                        cutpoints_label_bool=cutpoints_label_bool,
                        percentiles=percentiles,
                    )

        if synthetic_data is not None:
            instance_np = instance.detach().numpy()[0]
            y = np.array(
                synthetic_data(500000, 0, specific_instance=instance_np)[1].values
            )
            density = gaussian_kde(y)

            # Plotting
            x = np.linspace(x_range[0], x_range[1], num_interpolations)
            plt.plot(
                x,
                density(x),
                color="red",
                linewidth=2,
                alpha=alpha,
                label="True (KDE)",
            )

        if plot_mean_adjustment:
            plt.ylim(bottom=0)
            DP_glm = self.glm.mean(instance).item()
            DP_drn = self.drn.distributions(instance).mean.item()

            axes.axvline(
                DP_glm,
                ls="dashdot",
                alpha=0.9,
                linewidth=4,
                label=f"Baseline Mean:\n {round(float(DP_glm), 3)}",
                color="black",
            )
            axes.axvline(
                DP_drn,
                ls="dashdot",
                alpha=0.9,
                linewidth=4,
                label=f"Adjusted Mean:\n {round(float(DP_drn), 3)}",
                color="blue",
            )

        if synthetic_data is not None and plot_mean_adjustment:
            axes.axvline(
                np.mean(y),
                ls="dashdot",
                alpha=0.9,
                linewidth=4,
                label=f"True Mean:\n {round(float(np.mean(y)))}",
                color="red",
            )

        if observation is not None and synthetic_data is None:
            axes.axvline(
                observation,
                ls="dashed",
                alpha=0.75,
                linewidth=4,
                label=f"Observation:\n{round(float(observation), 3)}",
                color="red",
            )

        # Adding labels and title
        plt.xlabel("$y$")

        y_label = "$f(y|X=x^*)$"
        plt.gca().set_ylabel("")
        if plot_y_label is not None:
            y_label = plot_y_label
            plt.ylabel(y_label)

        title_name = (
            "Density Plot with Adjustment Factors"
            if plot_adjustments_labels
            else "Density Plot"
        )
        if plot_title is not None:
            title_name = plot_title
        plt.title(title_name)

        plt.legend(loc="upper right")
        if x_range is not None:
            x_min, x_max = x_range
        axes.set_xlim(x_min, x_max)
        y_max = y_max if plot_adjustments_labels else None
        if y_range is not None:
            y_max = y_range[1]
        axes.set_ylim(0, y_max)

    def cdf_plot(
        self,
        instance,
        grid=None,
        cutpoints=None,
        other_df_models=None,
        model_names=None,
        synthetic_data=None,
        x_range=None,
        plot_title=None,
        plot_baseline=True,
        density_transparency=1.0,
        dist_property="Mean",
        quantile_bounds=None,
        nsamples_background_fraction=0.1,
        adjustment=True,
        method="Kernel",
        labelling_gap=0.01,
        top_K_features=3,
        y_range=None,
        shap_fontsize=25,
        figsize=None,
    ):
        """
        Plot the cumulative distribution function.
        """
        alpha = density_transparency

        # Default cutpoints and num_interpolations setup
        cutpoints = self.default_cutpoints if cutpoints is None else cutpoints
        instance_raw = instance
        instance = (
            self._to_tensor(self.one_hot_encoder(instance))
            if self.all_categories is not None
            else self._to_tensor(instance)
        )
        # Interpolation
        lower_cutpoint = x_range[0] if x_range is not None else cutpoints[0]
        upper_cutpoint = x_range[1] if x_range is not None else cutpoints[-1]
        grid = (
            torch.linspace(lower_cutpoint, upper_cutpoint, 100).unsqueeze(-1)
            if grid is None
            else grid
        )

        # Assuming 'cdf_drn' computation is done as you've mentioned
        cdfs_glm = self.glm.distributions(instance).cdf(grid).detach().numpy()
        cdfs_drn = self.drn.distributions(instance).cdf(grid).detach().numpy()

        # Plotting
        figure, axes = plt.subplots(1, 1, figsize=figsize)
        if axes is not None and figsize is not None:
            axes.figure.set_size_inches(figsize)
        if axes is None:
            _, axes = plt.subplots(1, 1, figsize=figsize)

        plt.sca(axes)

        y_min = y_range[0] if y_range is not None else 1.0
        y_max = y_range[1] if y_range is not None else 0.0
        x_min, x_max = self.kernel_shap_plot(
            instance_raw=instance_raw,
            instance=instance,
            dist_property=dist_property,
            quantile_bounds=quantile_bounds,
            method=method,
            nsamples_background_fraction=nsamples_background_fraction,
            adjustment=adjustment,
            axes=axes,
            top_K_features=top_K_features,
            y_max=y_max,
            y_min=y_min,
            labelling_gap=labelling_gap,
            fontsize=shap_fontsize,
        )

        # Other DF Models Plotting
        if other_df_models is not None:
            for idx, DF_model in enumerate(other_df_models):
                current_cdf = (
                    DF_model.distributions(instance).cdf(grid).detach().numpy()
                )
                plt.plot(
                    grid,
                    current_cdf,
                    label=f"{model_names[idx]} CDF",
                    color=str(0.5 + 0.5 * idx / len(other_df_models)),
                    alpha=alpha,
                )

        # cdfs_mdn = mdn.distributions(torch.Tensor([0.5, 0.5]).reshape(1,2)).cdf(grid).detach().numpy()
        if synthetic_data is not None:
            samples = synthetic_data(
                n=500000, seed=1, specific_instance=instance.detach().numpy()[0]
            )[1].values
            if dist_property == "Mean":
                axes.axvline(
                    np.mean(samples),
                    ls="dashed",
                    alpha=0.75,
                    linewidth=4,
                    label=f"True {dist_property}:{round(float(np.mean(samples)), 3)}",
                    color="orange",
                )
            else:
                axes.axvline(
                    np.percentile(samples, 90),
                    ls="dashed",
                    alpha=0.75,
                    linewidth=4,
                    label=f"True {dist_property}:{round(float(np.percentile(samples, 90)), 3)}",
                    color="orange",
                )

        plt.plot(grid, cdfs_drn, label="DRN CDF", color="blue", linewidth=3)
        if plot_baseline:
            plt.plot(grid, cdfs_glm, label="GLM CDF", color="black", linewidth=3)
        if synthetic_data is not None:
            cdf_empirical = [
                self.empirical_cdf(samples, x) for x in grid.detach().numpy()
            ]
            plt.plot(grid, cdf_empirical, label="True CDF", color="orange", linewidth=3)

        x_min = x_range[0] if x_range is not None else x_min
        x_max = x_range[1] if x_range is not None else x_max
        axes.set_xlim(x_min, x_max)
        if y_range is not None:
            axes.set_ylim(y_range[0], y_range[1])
        plt.xlabel("$Y$")
        plt.ylabel("Cumulative Probability")

        plt.gca().set_ylabel("")
        if plot_title is not None:
            title_name = plot_title
            plt.title(title_name)

        # plt.title('Cumulative Distribution Functions (CDFs)', fontsize = 36)
        plt.legend()

    def kernel_shap_plot(
        self,
        instance_raw: pd.DataFrame,
        instance: torch.Tensor,
        dist_property: str,
        quantile_bounds: None,
        method="Kernel",
        nsamples_background_fraction: float = 1.0,
        adjustment: bool = True,
        axes=None,
        top_K_features: int = 3,
        # Plot styling parameters grouped at the end
        y_max: float = None,
        y_min: float = None,
        labelling_gap: float = 0.05,
        fontsize: int = 25,
    ):
        """
        Visualises the impact of SHAP values for the top K features on a specified distributional property,
        including the option to display adjustment effects.

        Parameters:
        - instance_raw (pd.DataFrame): The instance data prior to any processing.
        - instance (torch.Tensor): Processed instance data ready for the model.
        - dist_property (str): Target distributional property (e.g., 'Mean', 'Variance').
        - method: Method used for SHAP value computation, defaulting to 'Kernel'.
        - nsamples_background_fraction (float): Fraction of background data utilised, defaults to 1.0.
        - adjustment (bool): Whether to include adjustment effects in the visualisation, defaults to True.
        - axes: Matplotlib axes object for plotting.
        - top_K_features (int): Number of top features to highlight based on SHAP values.
        - Plot styling parameters like `y_max`, `y_min`, `labelling_gap`, `fontsize` are for visual adjustments.
        """

        # Compute SHAP values and model predictions
        if method == "Kernel":
            shap_base_value_diff, shap_values_diff, feature_names = self.kernel_shap(
                explaining_data=instance_raw,
                distributional_property=dist_property,
                nsamples_background_fraction=nsamples_background_fraction,
                adjustment=adjustment,
            ).shap_values_mean_adjustments()

            # Calculate distributional property values
            DP_glm, DP_drn = self._compute_distributional_properties(
                instance, dist_property, adjustment, quantile_bounds
            )
        else:
            # Placeholder for extending method functionality
            raise NotImplementedError(f"Method {method} is not implemented.")

        phi_j_s = np.concatenate((shap_base_value_diff, shap_values_diff[0]))

        # Get top K indices based on SHAP values
        top_K_indices = np.argsort(-np.abs(phi_j_s))[:top_K_features]

        # Extend feature names and instance values arrays
        feature_names = np.insert(feature_names, 0, "X_0")
        # instance_values = np.insert(instance.detach().numpy()[0], 0, 1.000)
        instance_values = np.insert(np.array(instance_raw)[0], 0, 1.000)

        # Select top K SHAP values, feature names, and instance values
        phi_j_s_top = phi_j_s[top_K_indices]
        feature_names_top = feature_names[top_K_indices]
        instance_values_top = instance_values[top_K_indices]

        # Create a boolean mask for all indices
        all_indices = np.arange(len(phi_j_s))
        mask = ~np.isin(all_indices, top_K_indices)

        # Get the remaining SHAP values
        phi_j_s_remaining = phi_j_s[mask]

        if adjustment:
            axes.axvline(
                DP_glm,
                ls="dashed",
                alpha=0.75,
                linewidth=4,
                label=f"Baseline {dist_property}:\n{round(float(DP_glm), 3)}",
                color="black",
            )

        # Plot vertical lines for means
        axes.axvline(
            DP_drn,
            ls="dashed",
            alpha=0.75,
            linewidth=4,
            label=f"DRN {dist_property}:\n{round(float(DP_drn), 3)}",
            color="blue",
        )

        # Determine x-axis range
        gap = abs(DP_glm - DP_drn) if adjustment else DP_drn / 2
        x_min, x_max = min(DP_glm, DP_drn) - 2 * gap, max(DP_glm, DP_drn) + 2 * gap

        # Plot SHAP value arrows and text
        y_position = y_max / 5
        if y_min is not None:
            y_position = y_min + (y_max - y_min) / 5
            y_max = y_max - y_min

        # For top K features
        for idx, phi in enumerate(phi_j_s_top):
            color = "green" if phi > 0 else "red"
            phi_sum_old = DP_glm + np.sum(phi_j_s_top[:idx])
            phi_sum_next = DP_glm + np.sum(phi_j_s_top[: (idx + 1)])

            y_position += y_max * labelling_gap

            if idx < (len(phi_j_s_top) - 1):
                plt.plot(
                    [phi_sum_next, phi_sum_next],
                    [y_position, y_position + y_max * labelling_gap],
                    ls="--",
                    color=color,
                    alpha=0.8,
                )
            # Draw the arrows
            axes.arrow(
                phi_sum_old,
                y_position,
                phi,
                0,
                head_width=y_max / 20,
                head_length=gap / 15,
                fc=color,
                ec=color,
                length_includes_head=True,
            )

            # Text the SHAP value
            instance_value = instance_values_top[idx]

            if not isinstance(instance_value, float):
                feature_name_text = str(instance_value)
            else:
                feature_name_text = f"{round(float(instance_value), 3)} "
            plt.text(
                phi_sum_old,
                y_position + (y_max / (50)),
                f"$\\phi_{{{top_K_indices[idx]}}}={round(float(phi), 3)}$; {feature_names_top[idx]}$= {feature_name_text}$",
                ha="left",
                fontsize=fontsize,
                color=color,
            )

            # Adjust the regions
            x_max = phi_sum_next + 2 * gap if phi_sum_next > x_max else x_max
            x_min = phi_sum_next - 2 * gap if phi_sum_next < x_min else x_min

        # Remaining features
        if len(phi_j_s_remaining) > 0:
            phi_remaining_sum = np.sum(phi_j_s_remaining)

            color = "green" if phi_remaining_sum > 0 else "red"

            plt.plot(
                [phi_sum_next, phi_sum_next],
                [y_position, y_position + y_max * labelling_gap],
                ls="--",
                color=color,
                alpha=0.8,
            )

            y_position += y_max * labelling_gap

            axes.arrow(
                phi_sum_next,
                y_position,
                phi_remaining_sum,
                0,
                head_width=y_max / 20,
                head_length=gap / 15,
                fc=color,
                ec=color,
                length_includes_head=True,
            )
            plt.text(
                phi_sum_next,
                y_position + (y_max / (50)),
                f"Remaining Features Contribute {round(float(phi_remaining_sum), 3)}",
                ha="left",
                fontsize=fontsize,
                color=color,
            )

        return (x_min, x_max)

    def _compute_distributional_properties(
        self, instance, dist_property, adjustment, quantile_bounds
    ):
        """
        Computes the specified distributional properties for both GLM and DRN models.
        """
        if dist_property == "Mean":
            DP_glm = self.glm.mean(instance).detach().numpy()[0] if adjustment else 0
            DP_drn = self.drn.distributions(instance).mean.detach().numpy()
        elif dist_property == "Variance":
            DP_glm = (
                self.glm.variance(instance).detach().numpy()[0] if adjustment else 0
            )
            DP_drn = self.drn.distributions(instance).variance().detach().numpy()
        else:
            percentile = (
                int(re.search(r"(\d+)% Quantile", dist_property).group(1))
                if re.search(r"(\d+)% Quantile", dist_property)
                else None
            )
            if quantile_bounds is not None:
                DP_glm = (
                    self.glm.quantiles(
                        instance,
                        [percentile],
                        l=quantile_bounds[0],
                        u=quantile_bounds[1],
                    ).item()
                    if percentile and adjustment
                    else 0
                )
                DP_drn = (
                    self.drn.distributions(instance)
                    .quantiles([percentile], l=quantile_bounds[0], u=quantile_bounds[1])
                    .item()
                    if percentile
                    else None
                )
            else:
                DP_glm = (
                    self.glm.quantiles(instance, [percentile]).item()
                    if percentile and adjustment
                    else 0
                )
                DP_drn = (
                    self.drn.distributions(instance).quantiles([percentile]).item()
                    if percentile
                    else None
                )
        return float(DP_glm), float(DP_drn)

    def empirical_cdf(self, samples, x):
        """
        Compute the empirical CDF for a given value x based on the provided samples.
        """
        return np.mean(samples <= x)

    def kernel_shap(
        self,
        explaining_data,
        distributional_property,
        adjustment=True,
        nsamples_background_fraction=1.0,
        glm_output=False,
        other_shap_values=None,
    ):
        """
        Pass on the explaining instance, background data, feature processing and value function to the KernelSHAP_DRN class
        """

        if distributional_property == "Mean":
            value_function = lambda instances: self.mean_value_function(
                instances, adjustment=adjustment
            )
            glm_value_function = self.mean_glm
        elif distributional_property == "Variance":
            value_function = lambda instances: self.variance_value_function(
                instances, adjustment=adjustment
            )
            glm_value_function = self.variance_glm
        else:
            value_function = self.set_value_function(
                distributional_property, adjustment, self.quantile_value_function
            )
            # glm_value_function = self.set_value_function(distributional_property, adjustment, self.quantile_glm)
            match = re.search(r"(\d+)% Quantile", distributional_property)
            if match:
                # Extract the percentage as an integer
                percentile = int(match.group(1))
                glm_value_function = lambda instances: self.quantile_glm(
                    instances, percentile=[percentile]
                )

            else:
                raise ValueError(
                    "Invalid distributional property! What Are We Explaining?"
                )

        return KernelSHAP_DRN(
            explaining_data,
            nsamples_background_fraction,
            self.background_data_df_before_one_hot,
            self.one_hot_encoder,
            value_function,
            glm_value_function if glm_output else None,
            other_shap_values,
        )

    def max_pdf_in_region(self, drn_pdf, glm_pdf, interval_width, cutpoint_idx):
        """
        Find the maximum pdf value within the region
        """
        drn_pdf_value = torch.max(
            torch.Tensor(
                drn_pdf[
                    cutpoint_idx * interval_width : (cutpoint_idx + 1) * interval_width
                ]
            )
        )
        glm_pdf_value = torch.max(
            torch.Tensor(
                glm_pdf[
                    cutpoint_idx * interval_width : (cutpoint_idx + 1) * interval_width
                ]
            )
        )
        max_of_two = torch.max(drn_pdf_value, glm_pdf_value)

        return max_of_two

    def region_adjustments(self, instance, region_start, region_end):
        """
        Calculate and round the adjustment factors
        """
        region_end = (
            region_end - 1e-03
            if region_end == self.default_cutpoints[-1]
            else region_end
        )
        factor = (
            self.real_adjustment_factors(instance, [region_start, region_end])
            .detach()
            .numpy()
        )
        factor_background = np.mean(
            self.real_adjustment_factors(
                self.background_data_tensor, [region_start, region_end]
            )
            .detach()
            .numpy(),
            axis=1,
        )
        rounded_factor = round(float(factor[0][0]), 2)
        rounded_factor_background = round(float(factor_background[0]), 2)

        return (rounded_factor, rounded_factor_background)

    def region_text(
        self,
        instance,
        interval_width,
        drn_pdf,
        glm_pdf,
        y_max,
        region_start,
        region_end,
        cutpoint_idx,
        adjustment_idx,
        cutpoints_label_bool=False,
        percentiles=None,
    ):
        """
        Text the density adjustment regions
        """
        max_of_two = self.max_pdf_in_region(
            drn_pdf, glm_pdf, interval_width, cutpoint_idx
        )
        rounded_factor, rounded_factor_background = self.region_adjustments(
            instance, region_start, region_end
        )

        if region_start != region_end:
            text_label = f"$\\hat{{a}}_{{{adjustment_idx}}}={rounded_factor}$"
            # text_label_background = f'Avg.$={rounded_factor_background}$'
            plt.text(
                float(region_start + region_end) / 2,
                max_of_two + y_max / 15,
                text_label,
                ha="center",
                fontsize=20,
                color="blue",
            )
            # plt.text(float(region_start + region_end) / 2,  max_of_two-y_max/15, text_label_background, fontsize=16, color='gray')

            adjustment_idx += 1

        if not cutpoints_label_bool:
            plt.axvline(region_end, ls="--", color="black", alpha=0.5)

        return adjustment_idx

    def real_adjustment_factors(self, instances, cutpoints) -> torch.tensor:
        """
        Calculate the real adjustment factors.
        """
        instances = self._to_tensor(instances)
        cutpoints = self._to_tensor(cutpoints)

        cdfs_baseline = self.glm.distributions(instances).cdf(cutpoints.unsqueeze(-1))
        cdfs_drn = self.drn.distributions(instances).cdf(cutpoints.unsqueeze(-1))

        prob_masses = torch.clamp(torch.diff(cdfs_drn, axis=0), min=1e-12)
        baseline_probs = torch.clamp(torch.diff(cdfs_baseline, axis=0), min=1e-12)
        epsilon = 1e-15
        real_adjustment_factors = (prob_masses + epsilon) / (baseline_probs + epsilon)

        return torch.clamp(real_adjustment_factors, min=0, max=10000)

    def _to_tensor(self, data):
        """
        Convert data to a torch.Tensor if not already.
        """
        if isinstance(data, torch.Tensor):
            return data
        return torch.tensor(
            np.asarray(data),
            dtype=next(self.glm.parameters()).dtype,
            device=next(self.glm.parameters()).device,
        )

    def one_hot_encoder(self, instances):
        """
        Apply one-hot encoding to categorical features, considering all possible categories from self.all_categories
        """
        # if self.column_transformer is not None:
        #     self.column_transformer.fit_transform(self.background_data_df_before_one_hot)
        #     instances = (self.column_transformer.transform(pd.DataFrame(instances,\
        #                      columns = self.background_data_df_before_one_hot.columns)))

        instances = pd.DataFrame(
            instances, columns=self.background_data_df_before_one_hot.columns
        )

        if self.cat_features:
            for feature in self.cat_features:
                instances[feature] = pd.Categorical(
                    instances[feature], categories=self.all_categories[feature]
                )

            instances = pd.get_dummies(instances, columns=self.cat_features).astype(
                float
            )

        if self.column_transformer is not None:
            background_df_one_hot = pd.get_dummies(
                self.background_data_df_before_one_hot, columns=self.cat_features
            ).astype(float)
            self.column_transformer.fit_transform(background_df_one_hot)
            instances = self.column_transformer.transform(instances)
            return instances

        else:
            return instances.values
        #     instances = pd.DataFrame(instances, columns = background_df_one_hot.columns)
        # else:
        #     instances = pd.DataFrame(instances, columns = self.background_data_df_before_one_hot.columns)

        # return(instances.values)

    def mean_drn(self, instances):
        """
        Calculate the mean predicted by the DRN network given the selected instances/features
        """
        instances = (
            self._to_tensor(self.one_hot_encoder(instances))
            if self.all_categories is not None
            else self._to_tensor(instances)
        )

        return self.drn.distributions(instances).mean.detach().numpy()

    def mean_glm(self, instances):
        """
        Calculate the mean predicted by the GLM given the selected instances/features
        """
        instances = (
            self._to_tensor(self.one_hot_encoder(instances))
            if self.all_categories is not None
            else self._to_tensor(instances)
        )

        return self.glm.distributions(instances).mean.detach().numpy()

    def mean_value_function(self, instances, adjustment):
        """
        Calculate the mean value function given the selected instances/features
        """
        return (
            self.mean_drn(instances) - self.mean_glm(instances)
            if adjustment
            else self.mean_drn(instances)
        )

    def quantile_drn(self, instances, percentile=[90], grid=None):
        """
        Calculate the quantile predicted by the DRN network given the selected instances/features
        """
        grid = (
            torch.linspace(
                self.default_cutpoints[0], self.default_cutpoints[-1], 3000
            ).unsqueeze(-1)
            if grid is None
            else grid
        )

        instances = (
            self._to_tensor(self.one_hot_encoder(instances))
            if self.all_categories is not None
            else self._to_tensor(instances)
        )

        return (
            self.drn.distributions(instances)
            .quantiles(percentile, l=grid[0], u=grid[-1] * 1.5)
            .detach()
            .numpy()
        )

    def quantile_glm(self, instances, percentile=[90], grid=None):
        """
        Calculate the quantile predicted by the GLM given the selected instances/features
        """
        grid = (
            torch.linspace(
                self.default_cutpoints[0], self.default_cutpoints[-1], 3000
            ).unsqueeze(-1)
            if grid is None
            else grid
        )

        instances = (
            self._to_tensor(self.one_hot_encoder(instances))
            if self.all_categories is not None
            else self._to_tensor(instances)
        )
        return (
            self.glm.quantiles(instances, percentile, l=grid[0], u=grid[-1] * 1.5)
            .detach()
            .numpy()
            .reshape(
                instances.shape[0],
            )
        )

    def quantile_value_function(
        self, instances, adjustment, grid=None, percentile=[90]
    ):
        """
        Calculate the quantile value function given the selected instances/features
        """
        grid = (
            torch.linspace(
                self.default_cutpoints[0], self.default_cutpoints[-1], 3000
            ).unsqueeze(-1)
            if grid is None
            else grid
        )

        return (
            self.quantile_drn(instances, percentile, grid).reshape(
                instances.shape[0],
            )
            - self.quantile_glm(instances, percentile, grid).reshape(
                instances.shape[0],
            )
            if adjustment
            else self.quantile_drn(instances, percentile, grid).reshape(
                instances.shape[0],
            )
        )

    def variance_glm(self, instances):

        raise NotImplementedError(
            f"Variance calculation for the GLM yet to be implemented!"
        )

    def variance_drn(self, instances):

        raise NotImplementedError(
            f"Variance calculation for the DRN network yet to be implemented!"
        )

    def variance_value_function(self, instances, adjustment):

        raise NotImplementedError(
            f"Kernel SHAP for variance has yet to be implemented!"
        )

    def set_value_function(self, distributional_property, adjustment, model_function):
        """
        Calculate the numeric part from the distributional property XX% quantile.
        Set the value function accordingly.
        """
        match = re.search(r"(\d+)% Quantile", distributional_property)
        if match:
            # Extract the percentage as an integer
            percentile = int(match.group(1))
            value_function = lambda instances: model_function(
                instances, adjustment=adjustment, percentile=[percentile]
            )
        else:
            raise ValueError("Invalid distributional property! What Are We Explaining?")

        return value_function


def plot_drn_density(
    x_grid,
    drn_pdf,
    index_left,
    index_right,
    axes,
    alpha=1.0,
):
    for i in range(index_left, index_right - 1):
        # Plot the horizontal line with increased thickness
        xs = x_grid[i : i + 2]
        ys = np.full_like(xs, drn_pdf[i])
        lab = "DRN" if i == index_left else None

        axes.plot(xs, ys, color="blue", linewidth=2, alpha=alpha, label=lab)

        # Draw vertical dashed lines between the jumps
        if i < index_right - 2:  # Check to avoid adding a vertical line at the end
            axes.vlines(
                xs[-1],
                drn_pdf[i],
                drn_pdf[i + 1],
                color="blue",
                linestyles="dotted",
                alpha=alpha,
            )



from typing import List, Union

import numpy as np
import torch
import torch.nn as nn
from torch.distributions import Categorical
from torch.distributions.mixture_same_family import MixtureSameFamily


class MDN(nn.Module):
    """
    Mixture density network that can switch between gamma and Gaussian distribution components.
    The distributional forecasts are mixtures of `num_components` specified distributions.
    """

    def __init__(
        self,
        p: int,
        num_hidden_layers=2,
        num_components=5,
        hidden_size=100,
        dropout_rate=0.2,
        distribution="gamma",
    ):
        """
        Args:
            p: the number of features in the model.
            num_hidden_layers: the number of hidden layers in the network.
            num_components: the number of components in the mixture.
            hidden_size: the number of neurons in each hidden layer.
            distribution: the type of distribution for the MDN ('gamma' or 'gaussian').
        """
        super(MDN, self).__init__()
        self.p = p
        self.num_components = num_components
        self.distribution = distribution

        layers = [nn.Linear(p, hidden_size), nn.LeakyReLU(), nn.Dropout(dropout_rate)]
        for _ in range(num_hidden_layers - 1):
            layers += [
                nn.Linear(hidden_size, hidden_size),
                nn.LeakyReLU(),
                nn.Dropout(dropout_rate),
            ]
        self.hidden_layers = nn.Sequential(*layers)

        # Output layers for mixture parameters
        self.logits = nn.Linear(hidden_size, num_components)
        if distribution == "gamma":
            self.log_alpha = nn.Linear(hidden_size, num_components)
            self.log_beta = nn.Linear(hidden_size, num_components)
        elif distribution == "gaussian":
            self.mu = nn.Linear(hidden_size, num_components)
            self.pre_sigma = nn.Linear(hidden_size, num_components)
        else:
            raise ValueError("Unsupported distribution: {}".format(distribution))

    def forward(self, x: torch.Tensor) -> List[torch.Tensor]:
        """
        Calculate the parameters of the mixture components.
        Args:
            x: the input features (shape: (n, p))
        Returns:
            A list containing the mixture weights, and distribution-specific parameters.
        """
        x = self.hidden_layers(x)
        weights = torch.softmax(self.logits(x), dim=1)

        if self.distribution == "gamma":
            alphas = torch.exp(self.log_alpha(x))
            betas = torch.exp(self.log_beta(x))
            return [weights, alphas, betas]
        else:
            mus = self.mu(x)
            sigmas = nn.Softplus()(self.pre_sigma(x))  # Ensure sigma is positive
            return [weights, mus, sigmas]

    def distributions(self, x: torch.Tensor) -> MixtureSameFamily:
        """
        Create distributional forecasts for the given inputs.
        Args:
            x: the input features (shape: (n, p))
        Returns:
            the predicted mixture distributions.
        """
        params = self.forward(x)
        weights = params[0]
        mixture = Categorical(weights)

        if self.distribution == "gamma":
            components = torch.distributions.Gamma(params[1], params[2])
        else:
            components = torch.distributions.Normal(params[1], params[2])

        return MixtureSameFamily(mixture, components)

    def mean(self, x: Union[np.ndarray, torch.Tensor]) -> np.ndarray:
        """
        Calculate the predicted means for the given observations, depending on the mixture distribution.
        Args:
            x: the input features (shape: (n, p))
        Returns:
            the predicted means (shape: (n,))
        """
        if isinstance(x, np.ndarray):
            x = torch.Tensor(x)
        distributions = self.distributions(x)
        return distributions.mean.detach().numpy()

    def icdf(
        self,
        x: Union[np.ndarray, torch.Tensor],
        p,
        l=None,
        u=None,
        max_iter=1000,
        tolerance=1e-7,
    ) -> torch.Tensor:
        """
        Calculate the inverse CDF (quantiles) of the distribution for the given cumulative probability.

        Args:
            p: cumulative probability values at which to evaluate icdf
            l: lower bound for the quantile search
            u: upper bound for the quantile search
            max_iter: maximum number of iterations permitted for the quantile search
            tolerance: stopping criteria for the search (precision)

        Returns:
            A tensor of shape (1, batch_shape) containing the inverse CDF values.
        """

        if isinstance(x, np.ndarray):
            x = torch.Tensor(x)
        dists = self.distributions(x)
        num_observations = dists.cdf(torch.Tensor([1]).unsqueeze(-1)).shape[
            1
        ]  # Dummy call to cdf to determine the batch size
        percentiles_tensor = torch.full(
            (1, num_observations), fill_value=p, dtype=torch.float32
        )

        # Initialise matrices for the bounds
        lower_bounds = (
            l if l is not None else torch.Tensor([0])
        )  # self.cutpoints[0] - (self.cutpoints[-1]-self.cutpoints[0])
        upper_bounds = (
            u if u is not None else torch.Tensor([200])
        )  # Adjust max value as needed

        lower_bounds = lower_bounds.repeat(num_observations).reshape(
            1, num_observations
        )
        upper_bounds = upper_bounds.repeat(num_observations).reshape(
            1, num_observations
        )

        for _ in range(max_iter):
            mid_points = (lower_bounds + upper_bounds) / 2
            cdf_vals = dists.cdf(mid_points)

            # Update the bounds based on where the CDF values are relative to the target percentiles
            lower_update = cdf_vals < percentiles_tensor
            upper_update = ~lower_update
            lower_bounds = torch.where(lower_update, mid_points, lower_bounds)
            upper_bounds = torch.where(upper_update, mid_points, upper_bounds)

            # Check for convergence
            if torch.max(upper_bounds - lower_bounds) < tolerance:
                break

        # Use the midpoint between the final bounds as the quantile estimate
        quantiles = (lower_bounds + upper_bounds) / 2

        return quantiles

    def quantiles(
        self,
        x: torch.Tensor,
        percentiles: list,
        l=None,
        u=None,
        max_iter=1000,
        tolerance=1e-7,
    ) -> torch.Tensor:
        """
        Calculate the quantile values for the given observations and percentiles (cumulative probabilities * 100).
        """
        l = (
            torch.Tensor([0]) if l is None else l
        )  # self.cutpoints[0] - (self.cutpoints[-1]-self.cutpoints[0]) if l is None else l
        u = torch.Tensor([200]) if u is None else u
        quantiles = [
            self.icdf(x, torch.tensor(percentile / 100.0), l, u, max_iter, tolerance)
            for percentile in percentiles
        ]

        return torch.stack(quantiles, dim=1)[0]


def gamma_mdn_loss(out: list[torch.Tensor], y: torch.Tensor) -> torch.Tensor:
    """
    Calculate the negative log-likelihood loss for the mixture density network.
    Args:
        out: the mixture weights, shape parameters and rate parameters (all shape: (n, num_components))
        y: the observed values (shape: (n, 1))
    Returns:
        the negative log-likelihood loss (shape: (1,))
    """
    weights, alphas, betas = out
    dists = MixtureSameFamily(
        Categorical(weights),
        torch.distributions.Gamma(alphas, betas),
    )
    log_prob = dists.log_prob(y.squeeze())
    assert log_prob.ndim == 1
    return -log_prob.mean()


def gaussian_mdn_loss(out: list[torch.Tensor], y: torch.Tensor) -> torch.Tensor:
    """
    Calculate the negative log-likelihood loss for the mixture density network.
    Args:
        out: the mixture weights, shape parameters and rate parameters (all shape: (n, num_components))
        y: the observed values (shape: (n, 1))
    Returns:
        the negative log-likelihood loss (shape: (1,))
    """
    weights, mus, sigmas = out
    dists = MixtureSameFamily(
        Categorical(weights),
        torch.distributions.Normal(mus, sigmas),
    )
    log_prob = dists.log_prob(y.squeeze())
    assert log_prob.ndim == 1
    return -log_prob.mean()



import os

os.environ["CUDA_VISIBLE_DEVICES"] = ""

import numpy as np
import pandas as pd
from synthetic_dataset import generate_synthetic_data

from drn import *


def test_plot_adjustment_factors():

    X_train, Y_train, train_dataset, val_dataset = generate_synthetic_data()
    y_train = Y_train.cpu().numpy()

    x_train = pd.DataFrame(
        X_train.detach().cpu().numpy(),
        columns=[f"X_{i}" for i in range(X_train.shape[1])],
    )

    cutpoints = drn_cutpoints(0, 1, 0.1, y_train, 2)

    glm = GLM.from_statsmodels(X_train, Y_train, distribution="gamma")

    drn = DRN(X_train.shape[1], cutpoints, glm)
    train(
        drn,
        drn_loss,
        train_dataset,
        val_dataset,
        epochs=2,
    )

    drn_explainer = DRNExplainer(
        drn,
        glm,
        cutpoints,
        x_train,
        cat_features=[],
    )

    instance = pd.DataFrame(
        np.array([[0.0, 1.0, 2.0, 3.0]]), columns=["X_1", "X_2", "X_3", "X_4"]
    )

    drn_explainer.plot_adjustment_factors(
        instance,
        num_interpolations=3000,
        plot_adjustments_labels=False,
        plot_mean_adjustment=True,
    )

    instance = np.array([[0.0, 1.0, 2.0, 3.0]])

    drn_explainer.plot_adjustment_factors(
        instance,
        num_interpolations=3000,
        plot_adjustments_labels=False,
        plot_mean_adjustment=True,
    )

    drn_explainer.cdf_plot(
        instance,
        method="Kernel",
        nsamples_background_fraction=0.5,
        top_K_features=3,
        labelling_gap=0.15,
        dist_property="90% Quantile",
        x_range=(2.5, 4.5),
        y_range=(0.87, 0.93),
        density_transparency=0.9,
        adjustment=True,
        shap_fontsize=40,
        plot_title="90% Quantile Adjustment Explanation",
        figsize=(20, 20),
    )



from typing import Optional, Union

import numpy as np
import torch
import torch.nn as nn

import statsmodels.api as sm
from statsmodels.genmod.families import Gaussian, Gamma


class GLM(nn.Module):
    """
    A base PyTorch model representing a generalized linear model.
    This class is extended by specific distribution types like Gamma or Gaussian.
    """

    def __init__(self, p: int, distribution: str):
        """
        Args:
            p: the number of features in the model
            distribution: the type of GLM ('gamma' or 'gaussian')
        """
        if not distribution in ("gamma", "gaussian"):
            raise ValueError(f"Unsupported model type: {distribution}")

        super(GLM, self).__init__()
        self.p = p
        self.distribution = distribution
        self.linear = nn.Linear(p, 1)
        self.dispersion = nn.Parameter(torch.tensor(torch.nan), requires_grad=False)

    @staticmethod
    def from_statsmodels(
        X: Union[np.ndarray, torch.Tensor],
        y: Union[np.ndarray, torch.Tensor],
        distribution: str,
    ):
        """
        Fit a GLM using statsmodels and initialize a PyTorch GLM model with the fitted parameters.

        Args:
            X: The feature matrix.
            y: The target vector.
            distribution: The type of distribution ('gamma' or 'gaussian').

        Returns:
            An instance of the PyTorch GLM class with parameters initialized to those estimated by statsmodels.
        """
        p = X.shape[1]

        if isinstance(X, torch.Tensor):
            device = X.device
            X = X.detach().cpu().numpy()
            y = y.detach().cpu().numpy()
        else:
            device = None

        # Choose the correct family based on the distribution
        if distribution == "gamma":
            family = Gamma(link=sm.families.links.Log())
        elif distribution == "gaussian":
            family = Gaussian()

        # Fit the GLM model
        model = sm.GLM(y, sm.add_constant(X), family=family)
        results = model.fit()
        betas = results.params
        if not isinstance(results.params, np.ndarray):
            betas = np.asarray(betas)

        # Create a new PyTorch GLM instance
        torch_glm = GLM(p, distribution)
        torch_glm.linear.weight.data = torch.tensor(
            betas[1:], dtype=torch.float32
        ).unsqueeze(0)
        torch_glm.linear.bias.data = torch.tensor([betas[0]], dtype=torch.float32)

        # Set additional parameters if needed
        if distribution == "gamma":
            disp = results.scale.item()
        elif distribution == "gaussian":
            # Standard deviation is the square root of the scale parameter
            disp = (results.scale**0.5).item()
        torch_glm.dispersion = nn.Parameter(torch.tensor(disp), requires_grad=False)

        if device:
            torch_glm = torch_glm.to(device)

        return torch_glm

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.distribution == "gamma":
            out = torch.exp(self.linear(x)).squeeze(-1)
        else:
            out = self.linear(x).squeeze(-1)

        assert out.shape == torch.Size([x.shape[0]])
        return out

    def clone(self) -> "GLM":
        """
        Create an independent copy of the model.
        """
        glm = GLM(self.p, self.distribution)
        glm.load_state_dict(self.state_dict())
        return glm

    def distributions(
        self, x: torch.Tensor
    ) -> Union[torch.distributions.Gamma, torch.distributions.Normal]:
        """
        Create distributional forecasts for the given inputs, specific to the model type.
        """
        if torch.isnan(self.dispersion):
            raise RuntimeError("Dispersion parameter has not been estimated yet.")

        if self.distribution == "gamma":
            alphas, betas = gamma_convert_parameters(self.forward(x), self.dispersion)
            return torch.distributions.Gamma(alphas, betas)
        else:
            return torch.distributions.Normal(self.forward(x), self.dispersion)

    def update_dispersion(self, X: torch.Tensor, y: torch.Tensor) -> None:
        disp = estimate_dispersion(self.distribution, self.forward(X), y, self.p)
        self.dispersion = nn.Parameter(torch.tensor(disp), requires_grad=False)

    def mean(self, x: torch.Tensor) -> torch.Tensor:
        """
        Calculate the predicted means for the given observations.
        """
        return self.forward(x)

    def icdf(
        self,
        x: Union[np.ndarray, torch.Tensor],
        p,
        l=None,
        u=None,
        max_iter=1000,
        tolerance=1e-7,
    ) -> torch.Tensor:
        """
        Calculate the inverse CDF (quantiles) of the distribution for the given cumulative probability.

        Args:
            p: cumulative probability values at which to evaluate icdf
            l: lower bound for the quantile search
            u: upper bound for the quantile search
            max_iter: maximum number of iterations permitted for the quantile search
            tolerance: stopping criteria for the search (precision)

        Returns:
            A tensor of shape (1, batch_shape) containing the inverse CDF values.
        """

        if isinstance(x, np.ndarray):
            x = torch.Tensor(x)
        dists = self.distributions(x)
        num_observations = dists.cdf(torch.Tensor([1]).unsqueeze(-1)).shape[
            1
        ]  # Dummy call to cdf to determine the batch size
        percentiles_tensor = torch.full(
            (1, num_observations), fill_value=p, dtype=torch.float32
        )

        # Initialise matrices for the bounds
        lower_bounds = (
            l if l is not None else torch.Tensor([0])
        )  # self.cutpoints[0] - (self.cutpoints[-1]-self.cutpoints[0])
        upper_bounds = (
            u if u is not None else torch.Tensor([200])
        )  # Adjust max value as needed

        lower_bounds = lower_bounds.repeat(num_observations).reshape(
            1, num_observations
        )
        upper_bounds = upper_bounds.repeat(num_observations).reshape(
            1, num_observations
        )

        for _ in range(max_iter):
            mid_points = (lower_bounds + upper_bounds) / 2
            cdf_vals = dists.cdf(mid_points)

            # Update the bounds based on where the CDF values are relative to the target percentiles
            lower_update = cdf_vals < percentiles_tensor
            upper_update = ~lower_update
            lower_bounds = torch.where(lower_update, mid_points, lower_bounds)
            upper_bounds = torch.where(upper_update, mid_points, upper_bounds)

            # Check for convergence
            if torch.max(upper_bounds - lower_bounds) < tolerance:
                break

        # Use the midpoint between the final bounds as the quantile estimate
        quantiles = (lower_bounds + upper_bounds) / 2

        return quantiles

    def quantiles(
        self,
        x: torch.Tensor,
        percentiles: list,
        l=None,
        u=None,
        max_iter=1000,
        tolerance=1e-7,
    ) -> torch.Tensor:
        """
        Calculate the quantile values for the given observations and percentiles (cumulative probabilities * 100).
        """
        if self.distribution == "gamma":
            quantiles = [
                self.icdf(x, torch.tensor(percentile / 100), l, u, max_iter, tolerance)
                for percentile in percentiles
            ]
        elif self.distribution == "gaussian":
            quantiles = [
                self.icdf(x, torch.tensor(percentile / 100), l, u, max_iter, tolerance)
                for percentile in percentiles
            ]
        else:
            raise ValueError(f"Unsupported model type: {self.distribution}")
        return torch.stack(quantiles, dim=1)[0]

    def quantiles_old(
        self, x: torch.Tensor, percentiles: list, grid: torch.Tensor
    ) -> torch.Tensor:
        # Get the CDF values for each instance and cutpoint
        cdf_values = self.distributions(x).cdf(grid).detach().numpy()
        quantile_levels = torch.zeros((len(percentiles), x.shape[0]))

        # For each instance and each percentile, find the closest cutpoint index
        for i, percentile in enumerate(percentiles):
            quantile_value = percentile / 100
            abs_diff = np.abs(cdf_values - quantile_value)
            closest_idx = abs_diff.argmin(axis=0)
            quantile_levels[i, :] = torch.Tensor(
                [grid[idx] for idx in closest_idx]
            ).reshape(1, x.shape[0])

        return quantile_levels


def gamma_deviance_loss(y_pred: torch.Tensor, y_true: torch.Tensor) -> torch.Tensor:
    """
    Calculate the Tweedie deviance loss for the gamma distribution.
    Args:
        y_pred: the predicted values (shape: (n,))
        y_true: the observed values (shape: (n,))
    Returns:
        the deviance loss (shape: (,))
    """
    loss = 2 * (y_true / y_pred - torch.log(y_true / y_pred) - 1)
    return torch.mean(loss)


def gaussian_deviance_loss(y_pred: torch.Tensor, y_true: torch.Tensor) -> torch.Tensor:
    """
    Calculate the Normal deviance loss for the Gaussian distribution.
    Args:
        y_pred: the predicted values (shape: (n,))
        y_true: the observed values (shape: (n,))
    Returns:
        the deviance loss (shape: (,))
    """
    loss = (y_true - y_pred) ** 2
    return torch.mean(loss)


def gamma_estimate_dispersion(mu: torch.Tensor, y: torch.Tensor, p: int) -> float:
    """
    For a gamma GLM, the dispersion parameter is estimated using the method of moments.
    Args:
        mu: the predicted means for the gamma distributions (shape: (n, 1))
        y: the observed values (shape: (n, 1))
        p: the number of features (not including the intercept)
    """
    n = mu.shape[0]
    dof = n - (p + 1)
    return (torch.sum((y - mu) ** 2 / mu**2) / dof).item()


def gamma_convert_parameters(
    mu: torch.Tensor, phi: torch.Tensor
) -> tuple[torch.Tensor, torch.Tensor]:
    """
    Our models predict the mean of the gamma distribution, but we need the shape and rate parameters.
    This function converts the mean and dispersion parameter into the shape and rate parameters.
    Args:
        mu: the predicted means for the gamma distributions (shape: (n,))
        phi: the dispersion parameter
    Returns:
        alpha: the shape parameter (shape: (n,))
        beta: the rate parameter (shape: (n,))
    """
    beta = 1.0 / (mu * phi)
    alpha = (1.0 / phi) * torch.ones_like(beta)
    return alpha, beta


def gaussian_estimate_sigma(mu: torch.Tensor, y: torch.Tensor) -> float:
    """
    For a Gaussian GLM, the dispersion parameter is estimated using the method of moments.
    Args:
        mu: the predicted means for the Gaussian distributions (shape: (n, 1))
        y: the observed values (shape: (n, 1))
        p: the number of features
    """
    n = mu.shape[0]
    variance_estimate = torch.sum((y - mu) ** 2) / (n - 1)
    return (torch.sqrt(variance_estimate)).item()


def estimate_dispersion(distribution: str, mu: torch.Tensor, y: torch.Tensor, p: int):
    """
    Estimate the dispersion parameter for the given distribution.
    """
    if distribution == "gamma":
        return gamma_estimate_dispersion(mu, y, p)
    elif distribution == "gaussian":
        return gaussian_estimate_sigma(mu, y)
    else:
        raise ValueError(f"Unsupported distribution: {distribution}")



import os

# os.environ["CUDA_VISIBLE_DEVICES"] = ""
os.environ["OMP_NUM_THREADS"] = "1"
os.environ["OPENBLAS_NUM_THREADS"] = "1"

import numpy as np
import pandas as pd
import torch
import statsmodels.api as sm
from statsmodels.genmod.families import Gamma
from synthetic_dataset import generate_synthetic_data

from drn import *


def check_crps(model, X_train, Y_train, grid_size=3000):
    grid = torch.linspace(0, Y_train.max().item() * 1.1, grid_size).unsqueeze(-1)
    grid = grid.to(X_train.device)
    dists = model.distributions(X_train)
    cdfs = dists.cdf(grid)
    grid = grid.squeeze()
    crps_ = crps(Y_train, grid, cdfs)
    assert crps_.shape == Y_train.shape
    assert crps_.mean() > 0


def test_glm():
    print("\n\nTraining GLM\n")
    X_train, Y_train, train_dataset, val_dataset = generate_synthetic_data()

    torch.manual_seed(1)
    glm = GLM(X_train.shape[1], distribution="gamma")

    train(
        glm,
        gamma_deviance_loss,
        train_dataset,
        val_dataset,
        epochs=2,
    )

    glm.update_dispersion(X_train, Y_train)

    check_crps(glm, X_train, Y_train)


def test_glm_from_statsmodels():
    print("\n\nTraining GLM\n")
    X_train, Y_train, train_dataset, val_dataset = generate_synthetic_data()

    # Construct GLM given training data in torch tensors
    glm = GLM.from_statsmodels(X_train, Y_train, distribution="gamma")

    our_dispersion = gamma_estimate_dispersion(
        glm.forward(X_train), Y_train, X_train.shape[1]
    )

    assert np.isclose(our_dispersion, glm.dispersion.item())

    # Construct GLM given training data in numpy arrays
    glm = GLM.from_statsmodels(
        X_train.detach().cpu().numpy(),
        Y_train.detach().cpu().numpy(),
        distribution="gamma",
    )
    glm = glm.to(
        X_train.device
    )  # since 'from_statsmodels' didn't know this information
    our_dispersion = gamma_estimate_dispersion(
        glm.forward(X_train), Y_train, X_train.shape[1]
    )

    assert np.isclose(our_dispersion, glm.dispersion.item())

    # Check we avoid this 'iloc' warning when training on pandas data types
    X_df = pd.DataFrame(
        X_train.detach().cpu().numpy(),
        columns=[f"X_{i}" for i in range(X_train.shape[1])],
    )
    y_ser = pd.Series(Y_train.detach().cpu().numpy(), name="Y")
    glm = GLM.from_statsmodels(X_df, y_ser, distribution="gaussian")

    # Check that pandas objects from split_and_preprocess
    (
        x_train,
        x_val,
        x_test,
        y_train,
        y_val,
        y_test,
        x_train_raw,
        x_val_raw,
        x_test_raw,
        num_features,
        cat_features,
        all_categories,
        ct,
    ) = split_and_preprocess(
        X_df, y_ser, ["X_0", "X_1", "X_2", "X_3"], [], seed=42, num_standard=True
    )

    glm = GLM.from_statsmodels(x_train, y_train, distribution="gamma")
    glm = glm.to(X_train.device)

    # Check that statsmodels predictions are just the same as ours
    # Choose the correct family based on the distribution
    model = sm.GLM(
        y_train, sm.add_constant(x_train), family=Gamma(link=sm.families.links.Log())
    )
    results = model.fit()
    statsmodels_predictions = results.predict(sm.add_constant(x_train))

    X_train = torch.tensor(x_train.values, dtype=X_train.dtype, device=X_train.device)
    our_predictions = glm.forward(X_train).detach().cpu().numpy()

    assert np.allclose(statsmodels_predictions, our_predictions)


def test_cann():
    print("\n\nTraining CANN\n")
    X_train, Y_train, train_dataset, val_dataset = generate_synthetic_data()

    torch.manual_seed(2)
    glm = GLM(X_train.shape[1], distribution="gamma")
    train(
        glm,
        gamma_deviance_loss,
        train_dataset,
        val_dataset,
        epochs=2,
    )

    cann = CANN(glm, num_hidden_layers=2, hidden_size=100)
    train(
        cann,
        gamma_deviance_loss,
        train_dataset,
        val_dataset,
        epochs=2,
    )

    cann.update_dispersion(X_train, Y_train)

    check_crps(cann, X_train, Y_train)


def test_mdn():
    print("\n\nTraining MDN\n")
    X_train, Y_train, train_dataset, val_dataset = generate_synthetic_data()

    torch.manual_seed(3)
    mdn = MDN(X_train.shape[1], num_components=5, distribution="gamma")
    train(
        mdn,
        gamma_mdn_loss,
        train_dataset,
        val_dataset,
        epochs=2,
    )

    check_crps(mdn, X_train, Y_train)


def setup_cutpoints(Y_train):
    max_y = torch.max(Y_train).item()
    len_y = torch.numel(Y_train)
    c_0 = 0.0
    c_K = max_y * 1.01
    p = 0.1
    num_cutpoints = int(np.ceil(p * len_y))
    cutpoints_ddr = list(np.linspace(c_0, c_K, num_cutpoints))
    assert len(cutpoints_ddr) >= 2
    return cutpoints_ddr


def test_ddr():
    print("\n\nTraining DDR\n")
    X_train, Y_train, train_dataset, val_dataset = generate_synthetic_data()

    cutpoints_ddr = setup_cutpoints(Y_train)

    torch.manual_seed(4)
    ddr = DDR(X_train.shape[1], cutpoints_ddr, hidden_size=100)
    train(ddr, ddr_loss, train_dataset, val_dataset, epochs=2)

    check_crps(ddr, X_train, Y_train)


def test_drn():
    print("\n\nTraining DRN\n")
    X_train, Y_train, train_dataset, val_dataset = generate_synthetic_data()
    y_train = Y_train.cpu().numpy()

    cutpoints_ddr = setup_cutpoints(Y_train)
    cutpoints_drn = merge_cutpoints(cutpoints_ddr, y_train, min_obs=2)
    assert len(cutpoints_drn) >= 2

    torch.manual_seed(5)
    glm = GLM(X_train.shape[1], distribution="gamma")
    train(
        glm,
        gamma_deviance_loss,
        train_dataset,
        val_dataset,
        epochs=2,
    )
    glm.update_dispersion(X_train, Y_train)

    drn = DRN(X_train.shape[1], cutpoints_drn, glm, hidden_size=100)

    # Try both loss functions with their regularisation terms enabled
    train(
        drn,
        lambda pred, y: drn_loss(
            pred, y, kind="jbce", kl_alpha=1, mean_alpha=1, dv_alpha=1, tv_alpha=1
        ),
        train_dataset,
        val_dataset,
        epochs=2,
    )
    train(
        drn,
        lambda pred, y: drn_loss(
            pred, y, kind="nll", kl_alpha=1, mean_alpha=1, dv_alpha=1, tv_alpha=1
        ),
        train_dataset,
        val_dataset,
        epochs=2,
    )

    check_crps(drn, X_train, Y_train)


def test_torch():
    X_train, Y_train, train_dataset, val_dataset = generate_synthetic_data()

    cutpoints = setup_cutpoints(Y_train)

    torch.manual_seed(5)
    glm = GLM.from_statsmodels(X_train, Y_train, distribution="gamma")

    hs = 5
    drn = DRN(X_train.shape[1], cutpoints, glm, num_hidden_layers=2, hidden_size=hs)
    train(
        drn,
        drn_loss,
        train_dataset,
        val_dataset,
        epochs=2,
    )

    # Calculate the expected number of weights & biases given two layers of hs hidden units
    expected_num_weights = hs * X_train.shape[1] + hs + hs * hs + hs
    num_weights = sum([p.numel() for p in drn.hidden_layers.parameters()])
    assert num_weights == expected_num_weights

    # Try again using np.int64 instead of ints for the hyperparameters
    drn = DRN(
        X_train.shape[1],
        cutpoints,
        glm,
        num_hidden_layers=np.int64(2),
        hidden_size=np.int64(hs),
    )
    num_weights = sum([p.numel() for p in drn.hidden_layers.parameters()])
    assert num_weights == expected_num_weights and len(drn.hidden_layers) // 3 == 2

    train(
        drn,
        drn_loss,
        train_dataset,
        val_dataset,
        epochs=2,
    )

    # Check dropout is working as intended
    drn = DRN(
        X_train.shape[1],
        cutpoints,
        glm,
        num_hidden_layers=2,
        hidden_size=hs,
        dropout_rate=0.5,
    )
    train(
        drn,
        drn_loss,
        train_dataset,
        val_dataset,
        epochs=2,
    )

    # Make sure two different predictions (which in drn.train mode) are different
    drn.train()
    preds1 = drn(X_train)[-1]
    preds2 = drn(X_train)[-1]
    assert not torch.allclose(preds1, preds2)

    # Make sure two different predictions (which in drn.eval mode) are the same
    drn.eval()
    preds1 = drn(X_train)[-1]
    preds2 = drn(X_train)[-1]
    assert torch.allclose(preds1, preds2)



import numpy as np
import torch
import torch.nn as nn

from ..distributions.histogram import Histogram


class DDR(nn.Module):
    def __init__(
        self, p: int, cutpoints, num_hidden_layers=2, hidden_size=100, dropout_rate=0.2
    ):
        """
        Args:
            x_train_shape: The shape of the training data, used to define the input size of the first layer.
            cutpoints: The cutpoints for the DDR model.
            num_hidden_layers: The number of hidden layers in the network.
            hidden_size: The number of neurons in each hidden layer.
        """
        super(DDR, self).__init__()
        self.cutpoints = nn.Parameter(torch.Tensor(cutpoints), requires_grad=False)
        self.p = p

        layers = [
            nn.Linear(self.p, hidden_size),
            nn.LeakyReLU(),
            nn.Dropout(dropout_rate),
        ]
        for _ in range(num_hidden_layers - 1):
            layers.append(nn.Linear(hidden_size, hidden_size))
            layers.append(nn.LeakyReLU())
            layers.append(nn.Dropout(dropout_rate))

        # Use nn.Sequential to chain the layers together
        self.hidden_layers = nn.Sequential(*layers)

        # Output layer for the pi values
        self.pi = nn.Linear(hidden_size, len(self.cutpoints) - 1)

    def forward(self, x):
        """
        Forward pass of the DDR model.
        Args:
            x: Input tensor.
        Returns:
            The cutpoints and probabilities for the DDR model.
        """
        # Pass input through the dynamically created hidden layers
        h = self.hidden_layers(x)

        # Calculate probabilities using the final layer
        probs = torch.softmax(self.pi(h), dim=1)

        return self.cutpoints, probs

    def distributions(self, x):
        cutpoints, prob_masses = self.forward(x)
        dists = Histogram(cutpoints, prob_masses)
        assert dists.batch_shape == torch.Size([x.shape[0]])
        return dists


def jbce_loss(dists, y, alpha=0.0):
    """
    The joint binary cross entropy loss.
    Args:
        dists: the predicted distributions
        y: the observed values
        alpha: the penalty parameter
    """

    cutpoints = dists.cutpoints
    cdf_at_cutpoints = dists.cdf_at_cutpoints()

    assert cdf_at_cutpoints.shape == torch.Size([len(cutpoints), len(y)])

    n = y.shape[0]
    C = len(cutpoints)

    # The cross entropy loss can't accept 0s or 1s for the cumulative probabilities.
    epsilon = 1e-15
    cdf_at_cutpoints = cdf_at_cutpoints.clamp(epsilon, 1 - epsilon)

    # Change: C to C-1
    losses = torch.zeros(C - 1, n, device=y.device, dtype=y.dtype)

    for i in range(1, C):
        targets = (y <= cutpoints[i]).float()
        probs = cdf_at_cutpoints[i, :]
        losses[i - 1, :] = nn.functional.binary_cross_entropy(
            probs, targets, reduction="none"
        )

    return torch.mean(losses)


def ddr_loss(pred, y, alpha=0.0):
    cutpoints, prob_masses = pred
    dists = Histogram(cutpoints, prob_masses)
    return jbce_loss(dists, y, alpha)


def nll_loss(dists, y, alpha=0.0):
    losses = -(dists.log_prob(y))
    return torch.mean(losses)


def ddr_cutpoints(c_0, c_K, p, y):
    num_cutpoints = int(np.ceil(p * len(y)))
    cutpoints = list(np.linspace(c_0, c_K, num_cutpoints))

    return cutpoints



import numpy as np
import torch


def generate_synthetic_data(n=1000, seed=1):
    rng = np.random.default_rng(seed)
    x_all = rng.random(size=(n, 4))
    epsilon = rng.normal(0, 0.2, n)

    means = np.exp(
        0
        - 0.5 * x_all[:, 0]
        + 0.5 * x_all[:, 1]
        + np.sin(np.pi * x_all[:, 0])
        - np.sin(np.pi * np.log(x_all[:, 2] + 1))
        + np.cos(x_all[:, 1] * x_all[:, 2])
    ) + np.cos(x_all[:, 1])
    dispersion = 0.5

    y_lognormal = np.exp(rng.normal(means / 4, dispersion))
    y_gamma = rng.gamma(1 / dispersion, scale=dispersion * means / 4)

    y_all = y_gamma * 0.5 + y_lognormal * 0.5 + epsilon**2

    num_val = int(n * 0.2)
    num_test = int(n * 0.2)
    num_train = n - num_val - num_test

    x_train, y_train = x_all[:num_train], y_all[:num_train]
    x_val, y_val = (
        x_all[num_train : num_train + num_val],
        y_all[num_train : num_train + num_val],
    )

    if torch.cuda.is_available():
        device = torch.device("cuda:0")
    else:
        device = torch.device("cpu")
    print(f"Using device {device}")

    # Convert NumPy arrays to PyTorch tensors and move them to the specified device
    X_train = torch.tensor(x_train).float().to(device)
    Y_train = torch.tensor(y_train).float().to(device)
    X_val = torch.tensor(x_val).float().to(device)
    Y_val = torch.tensor(y_val).float().to(device)

    # Create Tensor datasets for training and validation
    train_dataset = torch.utils.data.TensorDataset(X_train, Y_train)
    val_dataset = torch.utils.data.TensorDataset(X_val, Y_val)

    return X_train, Y_train, train_dataset, val_dataset



# __init__.py for distributionalforecasting package
from .models import *
from .distributions import *
from .interpretability import DRNExplainer, KernelSHAP_DRN
from .train import train, split_and_preprocess
from .metrics import crps, quantile_score, quantile_losses, rmse



import torch


def crps(
    obs: torch.Tensor,
    grid: torch.Tensor,
    cdf_on_grid: torch.Tensor,
):
    """
    Compute CRPS using the provided grid and CDF values with PyTorch tensors.

    :param x: observed value(s) as a PyTorch tensor
    :param y_grid: tensor of grid points
    :param CDF_grid: tensor of corresponding CDF values or a 2D tensor where each column is a CDF
    :return: CRPS value(s) as a PyTorch tensor
    """

    # Ensure x and CDF_grid are at least 1D and 2D tensors, respectively
    obs = obs.unsqueeze(0) if obs.ndim == 0 else obs
    cdf_on_grid = cdf_on_grid.unsqueeze(1) if cdf_on_grid.ndim == 1 else cdf_on_grid
    cdf_on_grid = cdf_on_grid.T

    # Compute the difference between grid points (assuming uniform spacing)
    dy = grid[1] - grid[0]

    # Calculate the Heaviside step function values for each x and y_grid value
    heaviside_matrix = (grid >= obs.unsqueeze(1)).type(torch.float32)

    # Compute the CRPS values for each x and CDF_grid pair
    crps_values = torch.sum((cdf_on_grid - heaviside_matrix) ** 2, dim=1) * dy

    # If x was a scalar, return a scalar. Otherwise, return a tensor.
    return crps_values if crps_values.numel() > 1 else crps_values.item()


def quantile_score(y_true, y_pred, p):
    """
    Compute the quantile score for predictions at a specific quantile.

    :param y_true: Actual target values as a Pandas Series or PyTorch tensor.
    :param y_pred: Predicted target values as a numpy array or PyTorch tensor.
    :param p: The cumulative probability as a float
    :return: The quantile score as a PyTorch tensor.
    """
    # Ensure that y_true and y_pred are PyTorch tensors
    y_true = (
        torch.Tensor(y_true.values) if not isinstance(y_true, torch.Tensor) else y_true
    )
    y_pred = torch.Tensor(y_pred) if not isinstance(y_pred, torch.Tensor) else y_pred
    # Reshape y_pred to match y_true if necessary and compute the error
    e = y_true - y_pred.reshape(y_true.shape)
    # Compute the quantile score
    return torch.where(y_true >= y_pred, p * e, (1 - p) * -e).mean()


def quantile_losses(
    p,
    model,
    model_name,
    X,
    y,
    max_iter=1000,
    tolerance=5e-5,
    l=None,
    u=None,
    print_score=True,
):
    """
    Calculate and optionally print the quantile loss for the given data and model.

    :param p: The cumulative probability ntile as a float
    :param model: The trained model.
    :param model_name: The name of the trained model.
    :param X: Input features as a Pandas DataFrame or numpy array.
    :param y: True target values as a Pandas Series or numpy array.
    :param max_iter: The maximum number of iterations for the quantile search algorithm.
    :param tolerance: The tolerance for convergence of the the quantile search algorithm.
    :param l: The lower bound for the quantile search
    :param u: The upper bound for the quantile search
    :param print_score: A boolean indicating whether to print the score.
    :return: The quantile loss as a PyTorch tensor.
    """
    # Predict quantiles based on the model name
    if model_name in ["GLM", "MDN", "CANN"]:
        predicted_quantiles = model.quantiles(
            X, [p * 100], max_iter=max_iter, tolerance=tolerance, l=l, u=u
        )
    elif model_name in ["DDR", "DRN"]:
        predicted_quantiles = model.distributions(X).quantiles(
            [p * 100], max_iter=max_iter, tolerance=tolerance, l=l, u=u
        )

    # Compute the quantile score
    score = quantile_score(y, predicted_quantiles, p)

    # Print the score if requested
    if print_score:
        print(f"{model_name}: {score:.5f}")

    return score


def rmse(y, y_hat):
    """
    Compute the Root Mean Square Error (RMSE) between the true values and predictions.

    :param y: True target values. Can be a Pandas Series or a PyTorch tensor.
    :param y_hat: Predicted target values. Should be a PyTorch tensor.
    :return: The RMSE as a PyTorch tensor.
    """
    # Convert y to a PyTorch tensor if it is not already one
    y = torch.Tensor(y.values) if not isinstance(y, torch.Tensor) else y
    # Calculate the RMSE
    return torch.sqrt(torch.mean((y.squeeze() - y_hat.squeeze()) ** 2))



import numpy as np
import torch
import torch.nn as nn

from ..distributions.extended_histogram import ExtendedHistogram
from .ddr import nll_loss


class DRN(nn.Module):
    def __init__(
        self,
        num_features,
        cutpoints,
        glm,
        num_hidden_layers=2,
        hidden_size=75,
        dropout_rate=0.2,
        baseline_start=False,
    ):
        """
        Args:
            num_features: Number of features in the input dataset.
            cutpoints: Cutpoints for the DRN model.
            glm: A Generalized Linear Model (GLM) that DRN will adjust.
            num_hidden_layers: Number of hidden layers in the DRN network.
            hidden_size: Number of neurons in each hidden layer.
        """
        super(DRN, self).__init__()
        self.cutpoints = nn.Parameter(torch.Tensor(cutpoints), requires_grad=False)
        # Assuming glm.clone() is a method to clone the glm model; ensure glm has a clone method.
        self.glm = glm.clone() if hasattr(glm, "clone") else glm

        layers = [
            nn.Linear(num_features, hidden_size),
            nn.LeakyReLU(),
            nn.Dropout(dropout_rate),
        ]
        for _ in range(num_hidden_layers - 1):
            layers.append(nn.Linear(hidden_size, hidden_size))
            layers.append(nn.LeakyReLU())
            layers.append(nn.Dropout(dropout_rate))

        self.hidden_layers = nn.Sequential(*layers)

        # Output layer
        self.fc_output = nn.Linear(hidden_size, len(self.cutpoints) - 1)
        self.batch_norm = nn.BatchNorm1d(len(cutpoints) - 1)

        # Initialize weights and biases for fc_output to zero
        if baseline_start:
            nn.init.constant_(self.fc_output.weight, 0)
            nn.init.constant_(self.fc_output.bias, 0)

    def log_adjustments(self, x):
        """
        Estimates log adjustments using the neural network.
        Args:
            x: Input features.
        Returns:
            Log adjustments for the DRN model.
        """
        # Pass input through the hidden layers
        z = self.hidden_layers(x)
        # Compute log adjustments
        log_adjustments = self.fc_output(z)
        return log_adjustments

        # normalized_log_adjustments = self.batch_norm(log_adjustments)
        # return normalized_log_adjustments

    def forward(self, x):
        DEBUG = True
        if DEBUG:
            num_cutpoints = len(self.cutpoints)
            num_regions = len(self.cutpoints) - 1

        with torch.no_grad():
            baseline_dists = self.glm.distributions(x)

            baseline_cdfs = baseline_dists.cdf(self.cutpoints.unsqueeze(-1)).T
            if DEBUG:
                assert baseline_cdfs.shape == (x.shape[0], num_cutpoints)

            baseline_probs = torch.diff(baseline_cdfs, dim=1)
            if DEBUG:
                assert baseline_probs.shape == (x.shape[0], num_regions)

            # Sometimes the GLM probabilities are 0 simply due to numerical problems.
            # DRN cannot adjust regions with 0 probability, so we ensure 0's become
            # an incredibly small number just to avoid this issue.
            mass = torch.sum(baseline_probs, axis=1, keepdim=True)
            baseline_probs = torch.clip(baseline_probs, min=1e-10, max=1.0)
            baseline_probs = (
                baseline_probs / torch.sum(baseline_probs, axis=1, keepdim=True) * mass
            )

        drn_logits = torch.log(baseline_probs) + self.log_adjustments(x)
        drn_pmf = torch.softmax(drn_logits, dim=1)

        if DEBUG:
            assert drn_pmf.shape == (x.shape[0], num_regions)

            # Sometimes we get nan value in here. Otherwise, it should sum to 1.
            assert torch.isnan(drn_pmf).any() or torch.allclose(
                torch.sum(drn_pmf, axis=1),
                torch.ones(x.shape[0], device=x.device),
            )

        return baseline_dists, self.cutpoints, baseline_probs, drn_pmf

    def distributions(self, x):
        baseline_dists, cutpoints, baseline_probs, drn_pmf = self.forward(x)
        return ExtendedHistogram(baseline_dists, cutpoints, drn_pmf, baseline_probs)


def jbce_loss(dists, y, alpha=0.0):
    """
    The joint binary cross entropy loss.
    Args:
        dists: the predicted distributions
        y: the observed values
        alpha: the penalty parameter
    """

    cutpoints = dists.cutpoints
    cdf_at_cutpoints = dists.cdf_at_cutpoints()

    assert cdf_at_cutpoints.shape == torch.Size([len(cutpoints), len(y)])

    n = y.shape[0]
    C = len(cutpoints)

    # The cross entropy loss can't accept 0s or 1s for the cumulative probabilities.
    epsilon = 1e-15
    cdf_at_cutpoints = cdf_at_cutpoints.clamp(epsilon, 1 - epsilon)

    # Change: C to C-1
    losses = torch.zeros(C - 1, n, device=y.device, dtype=y.dtype)

    for i in range(1, C):
        targets = (y <= cutpoints[i]).float()
        probs = cdf_at_cutpoints[i, :]
        losses[i - 1, :] = nn.functional.binary_cross_entropy(
            probs, targets, reduction="none"
        )

    return torch.mean(losses)


def drn_loss(
    pred,
    y,
    kind="jbce",
    kl_alpha=0,
    mean_alpha=0,
    tv_alpha=0,
    dv_alpha=0,
    kl_direction="forwards",
):
    baseline_dists, cutpoints, baseline_probs, drn_pmf = pred
    dists = ExtendedHistogram(baseline_dists, cutpoints, drn_pmf, baseline_probs)

    if kind == "jbce":
        losses = jbce_loss(dists, y)
    else:
        losses = nll_loss(dists, y)

    a_i = dists.real_adjustments()
    b_i = baseline_probs

    if kl_alpha > 0:
        epsilon = 1e-30
        if kl_direction == "forwards":
            kl = -(torch.log(a_i + epsilon) * b_i)
        else:
            kl = torch.log(a_i + epsilon) * a_i * b_i
        losses += torch.mean(torch.sum(kl, axis=0)) * kl_alpha

    if mean_alpha > 0:
        losses += torch.mean((baseline_dists.mean - dists.mean) ** 2) * mean_alpha

    if tv_alpha > 0 or dv_alpha > 0:
        drn_density = a_i * b_i / torch.diff(cutpoints)
        first_diffs = torch.diff(drn_density, dim=1)

        if tv_alpha > 0:
            losses += torch.mean(torch.sum(torch.abs(first_diffs), dim=1)) * tv_alpha

        if dv_alpha > 0:
            second_diffs = torch.diff(first_diffs, dim=1)
            losses += torch.mean(torch.sum(second_diffs**2, dim=1)) * dv_alpha

    return losses


def uniform_cutpoints(c_0, c_K, p=None, y=None, num_cutpoints=None):
    if p is not None:
        num_cutpoints = int(np.ceil(p * len(y)))
        return list(np.linspace(c_0, c_K, num_cutpoints))
    else:
        return list(np.linspace(c_0, c_K, num_cutpoints))


def merge_cutpoints(cutpoints: list[float], y: np.ndarray, min_obs: int) -> list[float]:
    # Ensure cutpoints are sorted and unique to start with
    cutpoints = sorted(list(np.unique(cutpoints)))
    assert len(cutpoints) >= 2

    new_cutpoints = [cutpoints[0]]  # Start with the first cutpoint
    left = 0

    for right in range(1, len(cutpoints) - 1):
        num_in_region = np.sum((y >= cutpoints[left]) & (y < cutpoints[right]))
        num_after_region = np.sum((y >= cutpoints[right]) & (y < cutpoints[-1]))

        if num_in_region >= min_obs and num_after_region >= min_obs:
            new_cutpoints.append(cutpoints[right])
            left = right

    new_cutpoints.append(cutpoints[-1])  # End with the last cutpoint

    return new_cutpoints


def drn_cutpoints(c_0, c_K, p=None, y=None, min_obs=1, num_cutpoints=int(100)):
    if y is None:
        raise ValueError(
            "The argument 'y' cannot be None. It must be a numpy array of target values."
        )
    cutpoints = uniform_cutpoints(c_0, c_K, p, y, num_cutpoints)
    return merge_cutpoints(cutpoints, y, min_obs)



import numpy as np
import pandas as pd
import pytest
from drn import *
from synthetic_dataset import generate_synthetic_data


# Sample test data and scenarios we discussed
@pytest.mark.parametrize(
    "L_Raw, data_train, M, expected",
    [
        # Test Case 1: Small dataset with clear cutpoints
        (
            [0, 3, 6, 9],
            np.array([0.5, 2.5, 4.5, 6.5, 8.5]),
            2,
            [0, 3, 9],
        ),  # As derived from our correct implementation
        # Test Case 2: Problematic case for the old algorithm where the final bucket might have less than M observations
        (
            [0, 3, 7, 9],
            np.array([0.5, 2.5, 4.5, 6.5, 8.5]),
            2,
            [0, 3, 9],
        ),  # Adjusted expectation to fit the correct algorithm output
    ],
)
def test_merge_cutpoints(L_Raw, data_train, M, expected):
    assert merge_cutpoints(L_Raw, data_train, M) == expected


def test_split_preprocess():

    X_train, Y_train, train_dataset, val_dataset = generate_synthetic_data()

    # Check we avoid this 'iloc' warning when training on pandas data types
    X_df = pd.DataFrame(
        X_train.detach().cpu().numpy(),
        columns=[f"X_{i}" for i in range(X_train.shape[1])],
    )
    y_ser = pd.Series(Y_train.detach().cpu().numpy(), name="Y")

    # Check that pandas objects from split_and_preprocess
    (
        x_train,
        x_val,
        x_test,
        y_train,
        y_val,
        y_test,
        x_train_raw,
        x_val_raw,
        x_test_raw,
        num_features,
        cat_features,
        all_categories,
        ct,
    ) = split_and_preprocess(
        X_df, y_ser, ["X_0", "X_1", "X_2", "X_3"], [], seed=42, num_standard=True
    )



from typing import Optional, Union

import numpy as np
import torch
import torch.nn as nn

from .glm import GLM, gamma_convert_parameters, estimate_dispersion


class CANN(nn.Module):
    """
    The Combined Actuarial Neural Network (CANN) model adaptable for both gamma and Gaussian GLMs.
    """

    def __init__(
        self,
        glm: GLM,
        num_hidden_layers=2,
        hidden_size=50,
        dropout_rate=0.2,
        train_glm=False,
    ):
        """
        Args:
            glm: the GLM to use as the backbone, can be either Gamma or Gaussian GLM
            num_hidden_layers: the number of hidden layers in the neural network
            hidden_size: the number of neurons in each hidden layer
            train_glm: whether to retrain the GLM or not
        """
        if not glm.distribution in ("gamma", "gaussian"):
            raise ValueError(f"Unsupported model type: {glm.distribution}")

        super(CANN, self).__init__()

        self.p = glm.p
        self.glm = glm.clone()
        self.train_glm = train_glm
        self.distribution = glm.distribution
        self.dispersion = nn.Parameter(torch.tensor(torch.nan), requires_grad=False)

        layers = [
            nn.Linear(self.p, hidden_size),
            nn.LeakyReLU(),
            nn.Dropout(dropout_rate),
        ]
        for _ in range(num_hidden_layers - 1):
            layers.append(nn.Linear(hidden_size, hidden_size))
            layers.append(nn.LeakyReLU())
            layers.append(
                nn.Dropout(dropout_rate)
            )  # Add dropout after each LeakyReLU activation

        layers.append(nn.Linear(hidden_size, 1))
        self.nn_output_layer = nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Calculate the predicted outputs for the distributions.
        Args:
            x: the input features (shape: (n, p))
        Returns:
            the predicted outputs (shape: (n,))
        """
        if self.distribution == "gamma":
            out = torch.exp(self.glm.linear(x) + self.nn_output_layer(x))
        else:
            out = self.glm.linear(x) + self.nn_output_layer(x)

        out = out.squeeze(-1)
        assert out.shape == torch.Size([x.shape[0]])
        return out

    def distributions(
        self, x: torch.Tensor
    ) -> Union[torch.distributions.Gamma, torch.distributions.Normal]:
        """
        Create distributional forecasts for the given inputs, specific to the model type.
        """
        if torch.isnan(self.dispersion):
            raise RuntimeError("Dispersion parameter has not been estimated yet.")

        if self.distribution == "gamma":
            alphas, betas = gamma_convert_parameters(self.forward(x), self.dispersion)
            dists = torch.distributions.Gamma(alphas, betas)
        else:
            dists = torch.distributions.Normal(self.forward(x), self.dispersion)

        assert dists.batch_shape == torch.Size([x.shape[0]])
        return dists

    def update_dispersion(self, X: torch.Tensor, y: torch.Tensor) -> None:
        disp = estimate_dispersion(self.distribution, self.forward(X), y, self.p)
        self.dispersion = nn.Parameter(torch.tensor(disp), requires_grad=False)

    def mean(self, x: np.ndarray) -> np.ndarray:
        """
        Calculate the predicted means for the given observations, specific to the model type.
        """
        return self.forward(torch.Tensor(x)).detach().numpy().squeeze()

    def icdf(
        self,
        x: Union[np.ndarray, torch.Tensor],
        p,
        l=None,
        u=None,
        max_iter=1000,
        tolerance=1e-7,
    ) -> torch.Tensor:
        """
        Calculate the inverse CDF (quantiles) of the distribution for the given cumulative probability.

        Args:
            p: cumulative probability values at which to evaluate icdf
            l: lower bound for the quantile search
            u: upper bound for the quantile search
            max_iter: maximum number of iterations permitted for the quantile search
            tolerance: stopping criteria for the search (precision)

        Returns:
            A tensor of shape (1, batch_shape) containing the inverse CDF values.
        """

        if isinstance(x, np.ndarray):
            x = torch.Tensor(x)
        dists = self.distributions(x)

        num_observations = dists.cdf(torch.Tensor([1]).unsqueeze(-1)).shape[
            1
        ]  # Dummy call to cdf to determine the batch size
        percentiles_tensor = torch.full(
            (1, num_observations), fill_value=p, dtype=torch.float32
        )

        # Initialise matrices for the bounds
        lower_bounds = (
            l if l is not None else torch.Tensor([0])
        )  # self.cutpoints[0] - (self.cutpoints[-1]-self.cutpoints[0])
        upper_bounds = (
            u if u is not None else torch.Tensor([200])
        )  # Adjust max value as needed

        lower_bounds = lower_bounds.repeat(num_observations).reshape(
            1, num_observations
        )
        upper_bounds = upper_bounds.repeat(num_observations).reshape(
            1, num_observations
        )

        for _ in range(max_iter):
            mid_points = (lower_bounds + upper_bounds) / 2
            cdf_vals = dists.cdf(mid_points)

            # Update the bounds based on where the CDF values are relative to the target percentiles
            lower_update = cdf_vals < percentiles_tensor
            upper_update = ~lower_update
            lower_bounds = torch.where(lower_update, mid_points, lower_bounds)
            upper_bounds = torch.where(upper_update, mid_points, upper_bounds)

            # Check for convergence
            if torch.max(upper_bounds - lower_bounds) < tolerance:
                break

        # Use the midpoint between the final bounds as the quantile estimate
        quantiles = (lower_bounds + upper_bounds) / 2

        return quantiles

    def quantiles(
        self,
        x: torch.Tensor,
        percentiles: list,
        l=None,
        u=None,
        max_iter=1000,
        tolerance=1e-7,
    ) -> torch.Tensor:
        """
        Calculate the quantile values for the given observations and percentiles (cumulative probabilities * 100).
        """
        if self.distribution == "gamma":
            quantiles = [
                self.icdf(
                    x, torch.tensor(percentile / 100.0), l, u, max_iter, tolerance
                )
                for percentile in percentiles
            ]
        else:
            quantiles = [
                self.icdf(
                    x, torch.tensor(percentile / 100.0), l, u, max_iter, tolerance
                )
                for percentile in percentiles
            ]
        return torch.stack(quantiles, dim=1)[0]



from .glm import (
    GLM,
    gamma_deviance_loss,
    gaussian_deviance_loss,
    gamma_estimate_dispersion,
    gamma_convert_parameters,
    gaussian_estimate_sigma,
)
from .cann import CANN
from .mdn import MDN, gamma_mdn_loss, gaussian_mdn_loss
from .ddr import DDR, jbce_loss, ddr_loss, nll_loss, ddr_cutpoints
from .drn import (
    DRN,
    drn_loss,
    uniform_cutpoints,
    merge_cutpoints,
    drn_cutpoints,
)

__all__ = (
    [
        "GLM",
        "gamma_deviance_loss",
        "gaussian_deviance_loss",
        "gamma_estimate_dispersion",
        "gamma_convert_parameters",
        "gaussian_estimate_sigma",
    ]
    + ["CANN"]
    + ["MDN", "gamma_mdn_loss", "gaussian_mdn_loss"]
    + ["DDR", "jbce_loss", "ddr_loss", "nll_loss", "ddr_cutpoints"]
    + [
        "DRN",
        "drn_loss",
        "uniform_cutpoints",
        "merge_cutpoints",
        "drn_cutpoints",
    ]
)



import torch
from torch.distributions import Distribution


class Histogram(Distribution):
    """
    This class represents a histogram distribution.
    Basically, the distribution is a composite of uniform distributions over the bins.
    """

    def __init__(self, cutpoints: torch.Tensor, prob_masses: torch.Tensor):
        """
        Args:
            regions: the bin boundaries (shape: (K+1,))
            prob_masses: the probability for landing in each regions (shape: (n, K))
        """

        # Constructed regions T_k \in [c_k, c_{k+1}) for all k \in \{0, ..., K-1\}
        self.cutpoints = cutpoints
        self.num_regions = len(self.cutpoints) - 1

        # Predicted probabilities vector: (Pr(c_0<Y<c_1|x,w_{Histogram}),..., Pr(c_{K-1}<Y<c_K|x,w_{Histogram}))
        self.prob_masses = prob_masses
        assert torch.allclose(
            torch.sum(self.prob_masses, dim=1),
            torch.ones(self.prob_masses.shape[0], device=cutpoints.device),
        )
        assert self.prob_masses.shape[1] == self.num_regions

        # Compute the bin widths for later use, i.e., (T_1 = c_1 - c_0,..., T_K = c_K - c_{K-1})
        self.bin_widths = self.cutpoints[1:] - self.cutpoints[:-1]
        assert torch.all(self.bin_widths > 0)

        # Compute the PDF values using the probability masses and bin widths, i.e., normalising
        # Pr(Y\in T_k|x,w_{Histogram})/T_k * Pr(c_0<Y<c_K|x,w_{Baseline})
        self.prob_densities = self.prob_masses / self.bin_widths

        super(Histogram, self).__init__(
            batch_shape=torch.Size([prob_masses.shape[0]]), validate_args=False
        )

    def prob(self, value: torch.Tensor) -> torch.Tensor:
        """
        Calculate the probability densities of `values`.
        """
        orig_ndim = value.ndim

        # Ensure the last dimension of value matches the batch_shape
        if value.shape[-1] != self.batch_shape[0]:
            if value.ndim == 1:
                value = value.unsqueeze(-1)
            value = value.expand(-1, self.batch_shape[0])

        # Ensure value is 2D
        if value.ndim == 1:
            value = value.unsqueeze(0)

        # Initialize CDF to zeros with the same shape as value
        probabilities = torch.zeros_like(value)

        # Go through each observation vector and calculate pdf over all batch_size distributions
        for i in range(value.shape[0]):
            # Calculate the pdf for the `y` batch
            y = value[i, :]

            # Iterate over each bin
            for r in range(self.num_regions):
                in_bin = (y >= self.cutpoints[r]) & (y < self.cutpoints[r + 1])
                probabilities[i, in_bin] = self.prob_densities[in_bin, r]

        # If we added a leading dimension, remove it
        if orig_ndim == 1 and probabilities.ndim == 2:
            probabilities = probabilities.squeeze(0)

        return probabilities

    def log_prob(self, value: torch.Tensor) -> torch.Tensor:
        """
        Calculate the log probability densities of `values`.
        """
        return torch.log(self.prob(value))

    def cdf(self, value: torch.Tensor) -> torch.Tensor:
        """
        Calculate the cumulative distribution function for the given values.
        """
        orig_ndim = value.ndim
        # Ensure the last dimension of value matches the batch_shape
        if value.shape[-1] != self.batch_shape[0]:  # and value.shape[-1] == 1:
            return self.cdf_same_eval(value)

        # Ensure value is 2D
        if value.ndim == 1:
            value = value.unsqueeze(0)

        # Initialize CDF to zeros with the same shape as value
        cdf_values = torch.zeros_like(value)

        # Original cdf
        cumulative_cdf = torch.cumsum(self.prob_masses, dim=-1).T

        # Iterate over each bin
        for j in range(value.shape[0]):

            y = value[j, :]

            # Conditions now compare each element in y against the cutpoints
            # This assumes the intent is to apply the condition across all dimensions uniformly
            condition_above_last_cutpoint = y >= self.cutpoints[-1]
            condition_below_first_cutpoint = y <= self.cutpoints[0]

            # Since y is a vector, the assignment needs to respect the condition per element
            cdf_values[j, condition_above_last_cutpoint] = 1.0
            cdf_values[j, condition_below_first_cutpoint] = 0.0
            # print(condition_above_last_cutpoint, y, cdf_values)

            # Determine the index of the cutpoints
            y_expanded = y.unsqueeze(0)  # Now y has shape [1, n]
            cutpoints_below = self.cutpoints[:-1].unsqueeze(
                1
            )  # Now self.cutpoints[:-1] has shape [K, 1]
            cutpoints_above = self.cutpoints[1:].unsqueeze(1)

            # Perform comparison to determine bins
            comparison_result_below = y_expanded >= cutpoints_below  # [K, n]
            comparison_result_above = y_expanded < cutpoints_above  # [K, n]
            valid_bins = comparison_result_below & comparison_result_above  # [K, n]

            # Find the last valid bin for each element
            last_bin_idx = valid_bins.long().argmax(dim=0)  # [n]
            # Ensure elements outside the cutpoints are handled correctly
            below_min_mask = y_expanded.squeeze() <= self.cutpoints[0]
            above_max_mask = y_expanded.squeeze() >= self.cutpoints[-1]

            # Update last_bin_idx for values outside the cutpoints
            last_bin_idx[below_min_mask] = (
                0  # First bin for values below the minimum cutpoint
            )
            last_bin_idx[above_max_mask] = (
                len(self.cutpoints) - 2
            )  # Last valid bin index for values above the maximum cutpoint

            # Determine next_bin_idx based on last_bin_idx
            next_bin_idx = last_bin_idx + 1
            # Ensure next_bin_idx does not exceed the number of bins
            next_bin_idx[above_max_mask] = len(self.cutpoints) - 1

            # Initialize 'last_cdfs' and 'next_cdfs' with zeros and ones, respectively
            zeros = torch.zeros(size=(1, value.shape[1]))
            ones = torch.ones(size=(1, value.shape[1]))

            # Ensure y is properly shaped for broadcasting. If y is already [1, n], this step might be redundant
            y = y.reshape(1, -1)  # Ensure y has shape [1, n]

            # Reshape or select cutpoints for broadcasting
            # If comparing against a single cutpoint, ensure it's shaped for broadcasting
            cutpoint_for_comparison_lower = self.cutpoints[1].reshape(
                1, -1
            )  # Ensure cutpoint is shaped [1, 1] or similar for broadcasting
            cutpoint_for_comparison_upper = self.cutpoints[-2].reshape(1, -1)

            # Perform the comparison
            condition_last_cdfs = (
                y >= cutpoint_for_comparison_lower
            )  # This should now result in a shape [1, n]
            condition_next_cdfs = y < cutpoint_for_comparison_upper

            # Update 'last_cdfs' based on condition, considering 'cumulative_cdf' indexing
            # Ensure 'last_bin_idx' and 'next_bin_idx' are correctly broadcasted or indexed to match 'y's dimensions
            last_cdfs = torch.where(
                condition_last_cdfs,
                cumulative_cdf[
                    (last_bin_idx - 1).unsqueeze(0),
                    torch.arange(cumulative_cdf.shape[1]),
                ],
                zeros,
            )
            next_cdfs = torch.where(
                condition_next_cdfs,
                cumulative_cdf[
                    (next_bin_idx - 1).unsqueeze(0),
                    torch.arange(cumulative_cdf.shape[1]),
                ],
                ones,
            )

            last_bin_idx_expanded = last_bin_idx.unsqueeze(0).expand(y.shape[0], -1)
            next_bin_idx_expanded = next_bin_idx.unsqueeze(0).expand(y.shape[0], -1)

            # Determine cutpoints based on conditions
            cutpoint_low = torch.where(
                y < self.cutpoints[1].unsqueeze(0),
                self.cutpoints[0],
                self.cutpoints[last_bin_idx_expanded],
            )
            cutpoint_high = torch.where(
                y >= self.cutpoints[-2].unsqueeze(0),
                self.cutpoints[-1],
                self.cutpoints[next_bin_idx_expanded],
            )

            # Determine bin_width based on conditions
            bin_width = torch.where(
                y < self.cutpoints[1].unsqueeze(0),
                self.bin_widths[0],
                self.bin_widths[last_bin_idx_expanded],
            )

            # Compute bin_fraction for each feature across all observations
            bin_fraction = (y - cutpoint_low) / bin_width

            # Compute cdf_values using the previously obtained last_cdfs and next_cdfs for each feature across all observations
            # Ensure last_cdfs and next_cdfs are retrieved using the approach described in previous steps and have the correct shape
            cdf_values[j, :] = last_cdfs + (next_cdfs - last_cdfs) * bin_fraction

        cdf_values = torch.clamp(cdf_values, max=1.0, min=0.0)

        # If we added a leading dimension, remove it
        if orig_ndim == 1 and cdf_values.ndim == 2:
            cdf_values = cdf_values.squeeze(0)

        return cdf_values

    def cdf_same_eval(self, value: torch.Tensor) -> torch.Tensor:
        """
        Calculate the cumulative distribution function for the same value across the batch.
        """
        orig_ndim = value.ndim

        # Ensure the last dimension of value matches the batch_shape
        if value.shape[-1] != self.batch_shape[0]:
            if value.ndim == 1:
                value = value.unsqueeze(-1)
            value = value.expand(-1, self.batch_shape[0])

        # Ensure value is 2D
        if value.ndim == 1:
            value = value.unsqueeze(0)

        # Initialize CDF to zeros with the same shape as value
        cdf_values = torch.zeros_like(value)

        # Original cdf
        cumulative_cdf = torch.cumsum(self.prob_masses, dim=-1).T

        # Iterate over each bin
        for j in range(value.shape[0]):
            y = value[j, 0]
            if y >= self.cutpoints[-1]:
                cdf_values[j, :] = torch.ones(size=(1, value.shape[1]))
                continue

            if y <= self.cutpoints[0]:
                cdf_values[j, :] = torch.zeros(size=(1, value.shape[1]))
                continue

            # Determine the index of the cutpoints
            last_bin_idx = (
                (y >= self.cutpoints[:-1]) & (y < self.cutpoints[1:])
            ).nonzero(as_tuple=True)[0]
            next_bin_idx = last_bin_idx + 1

            # Set cdfs for the lower and upper bounds
            last_cdfs = (
                torch.zeros(size=(1, value.shape[1]), device=value.device)
                if y < self.cutpoints[1]
                else cumulative_cdf[last_bin_idx - 1, :]
            )
            next_cdfs = (
                torch.ones(size=(1, value.shape[1]), device=value.device)
                if y >= self.cutpoints[-2]
                else cumulative_cdf[next_bin_idx - 1, :]
            )

            # Compute cdf_values
            cutpoint_low = (
                self.cutpoints[0]
                if y < self.cutpoints[1]
                else self.cutpoints[last_bin_idx]
            )
            cutpoint_high = (
                self.cutpoints[-1]
                if y >= self.cutpoints[-2]
                else self.cutpoints[next_bin_idx]
            )
            bin_width = (
                self.bin_widths[0]
                if y < self.cutpoints[1]
                else self.bin_widths[last_bin_idx]
            )
            bin_fraction = (y - cutpoint_low) / bin_width
            cdf_values[j, :] = last_cdfs + (next_cdfs - last_cdfs) * bin_fraction

        cdf_values = torch.clamp(cdf_values, max=1.0, min=0.0)

        # If we added a leading dimension, remove it
        if orig_ndim == 1 and cdf_values.ndim == 2:
            cdf_values = cdf_values.squeeze(0)

        return cdf_values

    def cdf_at_cutpoints(self) -> torch.Tensor:
        """
        Calculate the cumulative distribution function at each cutpoint.
        """
        # Want to use cumsum but add a 0 to the beginning, torch.cumsum(self.prob_masses, dim=1)
        return torch.cat(
            [
                torch.zeros(
                    self.prob_masses.shape[0], 1, device=self.prob_masses.device
                ),
                torch.cumsum(self.prob_masses, dim=1),
            ],
            dim=1,
        ).T

    @property
    def mean(self) -> torch.Tensor:
        """
        Calculate the mean of the distribution.
        Returns:
            the mean (shape: (batch_shape,))
        """
        middle_of_bins = (self.cutpoints[1:] + self.cutpoints[:-1]) / 2
        return torch.sum(self.prob_masses * middle_of_bins, dim=1)

    def icdf(self, p, l=None, u=None, max_iter=1000, tolerance=1e-7) -> torch.Tensor:
        """
        Calculate the inverse CDF (quantiles) of the distribution for the given cumulative probability.

        Args:
            p: cumulative probability values at which to evaluate icdf
            l: lower bound for the quantile search
            u: upper bound for the quantile search
            max_iter: maximum number of iterations permitted for the quantile search
            tolerance: stopping criteria for the search (precision)

        Returns:
            A tensor of shape (1, batch_shape) containing the inverse CDF values.
        """

        num_observations = self.cdf(torch.Tensor([1]).unsqueeze(-1)).shape[
            1
        ]  # Dummy call to cdf to determine the batch size
        percentiles_tensor = torch.full(
            (1, num_observations), fill_value=p, dtype=torch.float32
        )

        # Initialise matrices for the bounds
        lower_bounds = (
            l if l is not None else torch.Tensor([0])
        )  # self.cutpoints[0] - (self.cutpoints[-1]-self.cutpoints[0])
        upper_bounds = (
            u
            if u is not None
            else self.cutpoints[-1] + (self.cutpoints[-1] - self.cutpoints[0])
        )  # Adjust max value as needed

        lower_bounds = lower_bounds.repeat(num_observations).reshape(
            1, num_observations
        )
        upper_bounds = upper_bounds.repeat(num_observations).reshape(
            1, num_observations
        )

        for _ in range(max_iter):
            mid_points = (lower_bounds + upper_bounds) / 2

            cdf_vals = self.cdf(mid_points)

            # Update the bounds based on where the CDF values are relative to the target percentiles
            lower_update = cdf_vals < percentiles_tensor
            upper_update = ~lower_update
            lower_bounds = torch.where(lower_update, mid_points, lower_bounds)
            upper_bounds = torch.where(upper_update, mid_points, upper_bounds)

            # Check for convergence
            if torch.max(upper_bounds - lower_bounds) < tolerance:
                break

        # Use the midpoint between the final bounds as the quantile estimate
        quantiles = (lower_bounds + upper_bounds) / 2

        return quantiles

    def quantiles(
        self, percentiles: list, l=None, u=None, max_iter=1000, tolerance=1e-7
    ) -> torch.Tensor:
        """
        Calculate the quantile values for the given observations and percentiles (cumulative probabilities * 100).
        """

        l = torch.Tensor(
            [0]
        )  # self.cutpoints[0] - (self.cutpoints[-1]-self.cutpoints[0]) if l is None else l
        u = (
            self.cutpoints[-1] + (self.cutpoints[-1] - self.cutpoints[0])
            if u is None
            else u
        )
        quantiles = [
            self.icdf(torch.tensor(percentile / 100.0), l, u, max_iter, tolerance)
            for percentile in percentiles
        ]

        return torch.stack(quantiles, dim=1)[0]



import torch
import torch.nn as nn
from typing import Optional
from torch.distributions import Distribution

from .histogram import Histogram


class ExtendedHistogram(Distribution):
    """
    This class represents a splicing of a supplied distribution with a histogram distribution.
    The histogram part is defined by K regions with boundaries -infty < c_0 < c_1 < ... < c_K < infty.
    The final density before c_0 & after c_K is the same as the original distribution.
    The density between c_k & c_{k+1} is defined by the histogram distribution.
    """

    def __init__(
        self,
        baseline: Distribution,
        cutpoints: torch.Tensor,
        pmf: torch.Tensor,
        baseline_probs: Optional[torch.Tensor] = None,
    ):
        """
        Args:
            baseline: the original distribution
            cutpoints: the bin boundaries (shape: (K+1,))
            pmf: the refined (cond.) probability for landing in each region (shape: (n, K))
            baseline_probs: the baseline's probability for landing in each region (shape: (n, K))
        """
        self.baseline = baseline
        self.cutpoints = cutpoints
        self.prob_masses = pmf
        self.baseline_probs = baseline_probs
        self.histogram = Histogram(cutpoints, pmf)
        self.scale_down_hist = baseline.cdf(cutpoints[-1]) - baseline.cdf(cutpoints[0])

        assert self.scale_down_hist.shape == torch.Size([self.histogram.batch_shape[0]])

        super(ExtendedHistogram, self).__init__(
            batch_shape=self.histogram.batch_shape, validate_args=False
        )

    def baseline_prob_between_cutpoints(self) -> torch.Tensor:
        """
        Calculate the baseline probability vector
        """
        if self.baseline_probs is None:
            baseline_cdfs = self.baseline.cdf(self.cutpoints.unsqueeze(-1)).T
            self.baseline_probs = torch.diff(baseline_cdfs, dim=1)

        return self.baseline_probs

    def real_adjustments(self) -> torch.Tensor:
        """
        Calculate the real adjustment factors a_k's
        """
        return self.prob_masses / self.baseline_prob_between_cutpoints()

    def prob(self, value: torch.Tensor) -> torch.Tensor:
        """
        Calculate the probability densities of `values`.
        """

        orig_ndim = value.ndim

        # Ensure the last dimension of value matches the batch_shape
        if value.shape[-1] != self.batch_shape[0]:
            if value.ndim == 1:
                value = value.unsqueeze(-1)
            value = value.expand(-1, self.batch_shape[0])

        # Ensure value is 2D
        if value.ndim == 1:
            value = value.unsqueeze(0)

        baseline_prob = torch.exp(self.baseline.log_prob(value))
        hist_prob = self.histogram.prob(value) * (self.scale_down_hist + 1e-10)

        in_hist = (value >= self.histogram.cutpoints[0]) & (
            value < self.histogram.cutpoints[-1]
        )
        in_baseline = ~in_hist

        probabilities = torch.zeros_like(baseline_prob)
        probabilities[in_baseline] = baseline_prob[in_baseline]
        probabilities[in_hist] = hist_prob[in_hist]

        return probabilities

    def log_prob(self, value: torch.Tensor) -> torch.Tensor:
        return torch.log(self.prob(value))

    def cdf(self, value: torch.Tensor) -> torch.Tensor:
        """
        Calculate the cumulative distribution function for the given values.
        """
        baseline_cdf = self.baseline.cdf(value)
        hist_cdf = self.histogram.cdf(value) * self.scale_down_hist
        in_hist = (value >= self.histogram.cutpoints[0]) & (
            value < self.histogram.cutpoints[-1]
        )
        in_hist = (
            in_hist.expand(value.shape[0], self.batch_shape[0])
            if in_hist.ndim > 1
            else in_hist
        )
        in_baseline = ~in_hist

        lower_cdf = self.baseline.cdf(self.histogram.cutpoints[0])
        cdf_values = torch.zeros_like(baseline_cdf)

        cdf_values[in_baseline] = baseline_cdf[in_baseline]
        cdf_values[in_hist] = (lower_cdf + hist_cdf)[in_hist]

        return cdf_values

    def cdf_at_cutpoints(self) -> torch.Tensor:
        """
        Calculate the cumulative distribution function at each cutpoint.
        """
        hist_at_cutpoints = (
            self.histogram.cdf_at_cutpoints() * self.scale_down_hist.unsqueeze(0)
        )
        lower_cdf = self.baseline.cdf(self.histogram.cutpoints[0]).unsqueeze(0)
        out = lower_cdf + hist_at_cutpoints
        return out

    @property
    def mean(self) -> torch.Tensor:
        """
        Calculate the mean of the distribution.
        Returns:
            the mean (shape: (batch_shape,))
        """
        middle_of_bins = (self.cutpoints[1:] + self.cutpoints[:-1]) / 2
        return torch.sum(self.prob_masses * middle_of_bins, dim=1)

    def icdf(self, p, l=None, u=None, max_iter=1000, tolerance=1e-7) -> torch.Tensor:
        """
        Calculate the inverse CDF (quantiles) of the distribution for the given cumulative probability.

        Args:
            p: cumulative probability values at which to evaluate icdf
            l: lower bound for the quantile search
            u: upper bound for the quantile search
            max_iter: maximum number of iterations permitted for the quantile search
            tolerance: stopping criteria for the search (precision)

        Returns:
            A tensor of shape (1, batch_shape) containing the inverse CDF values.
        """

        num_observations = self.cdf(torch.Tensor([1]).unsqueeze(-1)).shape[
            1
        ]  # Dummy call to cdf to determine the batch size
        percentiles_tensor = torch.full(
            (1, num_observations), fill_value=p, dtype=torch.float32
        )

        # Initialise matrices for the bounds
        lower_bounds = (
            l if l is not None else torch.Tensor([0])
        )  # self.cutpoints[0] - (self.cutpoints[-1]-self.cutpoints[0])
        upper_bounds = (
            u
            if u is not None
            else self.cutpoints[-1] + (self.cutpoints[-1] - self.cutpoints[0])
        )  # Adjust max value as needed

        lower_bounds = lower_bounds.repeat(num_observations).reshape(
            1, num_observations
        )
        upper_bounds = upper_bounds.repeat(num_observations).reshape(
            1, num_observations
        )

        for _ in range(max_iter):
            mid_points = (lower_bounds + upper_bounds) / 2

            cdf_vals = self.cdf(mid_points)

            # Update the bounds based on where the CDF values are relative to the target percentiles
            lower_update = cdf_vals < percentiles_tensor
            upper_update = ~lower_update
            lower_bounds = torch.where(lower_update, mid_points, lower_bounds)
            upper_bounds = torch.where(upper_update, mid_points, upper_bounds)

            # Check for convergence
            if torch.max(upper_bounds - lower_bounds) < tolerance:
                break

        # Use the midpoint between the final bounds as the quantile estimate
        quantiles = (lower_bounds + upper_bounds) / 2

        return quantiles

    def quantiles(
        self, percentiles: list, l=None, u=None, max_iter=1000, tolerance=1e-7
    ) -> torch.Tensor:
        """
        Calculate the quantile values for the given observations and percentiles (cumulative probabilities * 100).
        """
        l = torch.Tensor(
            self.cutpoints[0] - (self.cutpoints[-1] - self.cutpoints[0])
            if l is None
            else l
        )
        u = (
            self.cutpoints[-1] + (self.cutpoints[-1] - self.cutpoints[0])
            if u is None
            else u
        )
        quantiles = [
            self.icdf(torch.tensor(percentile / 100.0), l, u, max_iter, tolerance)
            for percentile in percentiles
        ]

        return torch.stack(quantiles, dim=1)[0]
